//############################################################################//
//Functions and Buffer Configurations
#define _FS_MINIMIZE 0 /* 0 to 3 */
/* The _FS_MINIMIZE option defines minimization level to remove API functions.
/
/   0: All basic functions are enabled.
/   1: f_stat(), f_getfree(), f_unlink(), f_mkdir(), f_chmod(), f_utime(),
/      f_truncate() and f_rename() function are removed.
/   2: f_opendir(), f_readdir() and f_closedir() are removed in addition to 1.
/   3: f_lseek() function is removed in addition to 2. */


#define _USE_FASTSEEK 0 /* 0:Disable or 1:Enable */
/* To enable fast seek feature, set _USE_FASTSEEK to 1. */

/*---------------------------------------------------------------------------/
/ Locale and Namespace Configurations
/----------------------------------------------------------------------------*/

#define _CODE_PAGE 1251
/* The _CODE_PAGE specifies the OEM code page to be used on the target system.
/  Incorrect setting of the code page can cause a file open failure.
/   1251 - Cyrillic (Windows)
/   866  - Russian (OEM)
/   1    - ASCII (Valid for only non-LFN cfg.)
*/


#define _USE_LFN 0  /* 0 to 3 */
#define _MAX_LFN 255  /* Maximum LFN length to handle (12 to 255) */
/* The _USE_LFN option switches the LFN feature.
/
/   0: Disable LFN feature. _MAX_LFN has no effect.
/   1: Enable LFN with static working buffer on the BSS. Always NOT reentrant.
/   2: Enable LFN with dynamic working buffer on the STACK.
/   3: Enable LFN with dynamic working buffer on the HEAP.
/
/  To enable LFN feature, Unicode handling functions ff_convert() and ff_wtoupper()
/  function must be added to the project.
/  The LFN working buffer occupies (_MAX_LFN + 1) * 2 bytes. When use stack for the
/  working buffer, take care on stack overflow. When use heap memory for the working
/  buffer, memory management functions, ff_memalloc() and ff_memfree(), must be added
/  to the project. */

/*---------------------------------------------------------------------------/
/ System Configurations
/----------------------------------------------------------------------------*/

#define _WORD_ACCESS 1 /* 0 or 1 */
/* The _WORD_ACCESS option is an only platform dependent option. It defines
/  which access method is used to the word data on the FAT volume.
/
/   0: Byte-by-byte access. Always compatible with all platforms.
/   1: Word access. Do not choose this unless under both the following conditions.
/
/  * Byte order on the memory is little-endian.
/  * Address miss-aligned word access is always allowed for all instructions.
/
/  If it is the case, _WORD_ACCESS can also be set to 1 to improve performance
/  and reduce code size.
*/

#define _FS_TIMEOUT  1000 /* Timeout period in unit of time ticks */




/* Definitions of volume management */

#define LD2PD(vol) (byte)(vol) /* Each logical drive is bound to the same physical drive number */


/* Type of path name strings on FatFs API */
#ifndef _INC_char
#define _T(x) x
#define _TEXT(x) x
#endif




/* File system object structure (FATFS) */

typedef struct {
 byte fs_type;  /* FAT sub-type (0:Not mounted) */
 byte drv;   /* Physical drive number */
 byte csize;   /* Sectors per cluster (1,2,4...128) */
 byte n_fats;   /* Number of FAT copies (1 or 2) */
 byte wflag;   /* win[] flag (b0:dirty) */
 byte fsi_flag;  /* FSINFO flags (b7:disabled, b0:dirty) */
 word id;    /* File system mount ID */
 word n_rootdir;  /* Number of root directory entries (FAT12/16) */
 dword last_clust;  /* Last allocated cluster */
 dword free_clust;  /* Number of free clusters */
 dword n_fatent;  /* Number of FAT entries (= number of clusters + 2) */
 dword fsize;   /* Sectors per FAT */
 dword volbase;  /* Volume start sector */
 dword fatbase;  /* FAT start sector */
 dword dirbase;  /* Root directory start sector (FAT32:Cluster#) */
 dword database;  /* Data start sector */
 dword winsect;  /* Current sector appearing in the win[] */
 byte win[512]; /* Disk access window for Directory, FAT (and file data at tiny cfg) */
} FATFS;



/* File object structure (FIL) */

typedef struct {
 FATFS* fs;    /* Pointer to the related file system object (**do not change order**) */
 word id;    /* Owner file system mount ID (**do not change order**) */
 byte flag;   /* File status flags */
 byte err;   /* Abort flag (error code) */
 dword fptr;   /* File read/write pointer (Zeroed on file open) */
 dword fsize;   /* File size */
 dword sclust;   /* File data start cluster (0:no data cluster, always 0 when fsize is 0) */
 dword clust;   /* Current cluster of fpter */
 dword dsect;   /* Current data sector of fpter */
 dword dir_sect;  /* Sector containing the directory entry */
 byte* dir_ptr;  /* Pointer to the directory entry in the window */
#if _USE_FASTSEEK
 dword* cltbl;   /* Pointer to the cluster link map table (Nulled on file open) */
#endif
} FIL;



/* Directory object structure (DIR) */

typedef struct {
 FATFS* fs;    /* Pointer to the owner file system object (**do not change order**) */
 word id;    /* Owner file system mount ID (**do not change order**) */
 word index;   /* Current read/write index number */
 dword sclust;   /* Table start cluster (0:Root dir) */
 dword clust;   /* Current cluster */
 dword sect;   /* Current sector */
 byte* dir;   /* Pointer to the current SFN entry in the win[] */
 byte* fn;    /* Pointer to the SFN (in/out) {file[8],ext[3],status[1]} */
#if _USE_LFN
 word* lfn;   /* Pointer to the LFN working buffer */
 word lfn_idx;  /* Last matched LFN index number (0xFFFF:No LFN) */
#endif
} DIR;



/* File status structure (FILINFO) */

typedef struct {
 dword fsize;   /* File size */
 word fdate;   /* Last modified date */
 word ftime;   /* Last modified time */
 byte fattrib;  /* Attribute */
 char fname[13];  /* Short file name (8.3 format) */
#if _USE_LFN
 char* lfname;   /* Pointer to the LFN buffer */
 word  lfsize;   /* Size of LFN buffer in char */
#endif
} FILINFO;



/* File function return code (FRESULT) */

typedef enum {
 FR_OK=0,    /* (0) Succeeded */
 FR_DISK_ERR,   /* (1) A hard error occurred in the low level disk I/O layer */
 FR_int_ERR,    /* (2) Assertion failed */
 FR_NOT_READY,   /* (3) The physical drive cannot work */
 FR_NO_FILE,    /* (4) Could not find the file */
 FR_NO_PATH,    /* (5) Could not find the path */
 FR_INVALID_NAME,  /* (6) The path name format is invalid */
 FR_DENIED,    /* (7) Access denied due to prohibited access or directory full */
 FR_EXIST,    /* (8) Access denied due to prohibited access */
 FR_INVALID_OBJECT,  /* (9) The file/directory object is invalid */
 FR_WRITE_PROTECTED,  /* (10) The physical drive is write protected */
 FR_INVALID_DRIVE,  /* (11) The logical drive number is invalid */
 FR_NOT_ENABLED,   /* (12) The volume has no work area */
 FR_NO_FILESYSTEM,  /* (13) There is no valid FAT volume */
 FR_MKFS_ABORTED,  /* (14) The f_mkfs() aborted due to any parameter error */
 FR_TIMEOUT,    /* (15) Could not get a grant to access the volume within defined period */
 FR_LOCKED,    /* (16) The operation is rejected according to the file sharing policy */
 FR_NOT_ENOUGH_CORE,  /* (17) LFN working buffer could not be allocated */
 FR_TOO_MANY_OPEN_FILES, /* (18) Number of open files > _FS_SHARE */
 FR_INVALID_PARAMETER /* (19) Given parameter is invalid */
} FRESULT;



/*--------------------------------------------------------------*/
/* FatFs module application interface                           */

FRESULT f_open (FIL* fp, const char* path, byte mode);    /* Open or create a file */
FRESULT f_close (FIL* fp);           /* Close an open file object */
FRESULT f_read (FIL* fp, void* buff, word btr, word* br);   /* Read data from a file */
FRESULT f_write (FIL* fp, const void* buff, word btw, word* bw); /* Write data to a file */
FRESULT f_forward (FIL* fp, word(*func)(const byte*,word), word btf, word* bf); /* Forward data to the stream */
FRESULT f_lseek (FIL* fp, dword ofs);        /* Move file pointer of a file object */
FRESULT f_truncate (FIL* fp);          /* Truncate file */
FRESULT f_sync (FIL* fp);           /* Flush cached data of a writing file */
FRESULT f_opendir (DIR* dp, const char* path);      /* Open a directory */
FRESULT f_closedir (DIR* dp);          /* Close an open directory */
FRESULT f_readdir (DIR* dp, FILINFO* fno);       /* Read a directory item */
FRESULT f_mkdir (const char* path);        /* Create a sub directory */
FRESULT f_unlink (const char* path);        /* Delete an existing file or directory */
FRESULT f_rename (const char* path_old, const char* path_new); /* Rename/Move a file or directory */
FRESULT f_stat (const char* path, FILINFO* fno);     /* Get file status */
FRESULT f_chmod (const char* path, byte value, byte mask);   /* Change attribute of the file/dir */
FRESULT f_utime (const char* path, const FILINFO* fno);   /* Change times-tamp of the file/dir */
FRESULT f_chdir (const char* path);        /* Change current directory */
FRESULT f_chdrive (const char* path);        /* Change current drive */
FRESULT f_getcwd (char* buff, word len);       /* Get current directory */
FRESULT f_getfree (dword* nclst, FATFS** fatfs); /* Get number of free clusters on the drive */
FRESULT f_getlabel (const char* path, char* label, dword* sn); /* Get volume label */
FRESULT f_setlabel (const char* label);       /* Set volume label */
FRESULT f_mkfs (const char* path, byte sfd, word au);    /* Create a file system on the volume */
FRESULT f_fdisk (byte pdrv, const dword szt[], void* work);   /* Divide a physical drive into some partitions */
int f_putc (char c, FIL* fp);          /* Put a character to the file */
int f_puts (const char* str, FIL* cp);        /* Put a string to the file */
int f_printf (FIL* fp, const char* str, ...);      /* Put a formatted string to the file */
char* f_gets (char* buff, int len, FIL* fp);      /* Get a string from the file */

#define f_eof(fp) (((fp)->fptr==(fp)->fsize) ? 1 : 0)
#define f_error(fp) ((fp)->err)
#define f_tell(fp) ((fp)->fptr)
#define f_size(fp) ((fp)->fsize)

#ifndef EOF
#define EOF (-1)
#endif




/*--------------------------------------------------------------*/
/* Additional user defined functions                            */

/* Unicode support functions */
#if _USE_LFN       /* Unicode - OEM code conversion */
word ff_convert (word chr, word dir); /* OEM-Unicode bidirectional conversion */
word ff_wtoupper (word chr);   /* Unicode upper-case conversion */
#if _USE_LFN==3      /* Memory functions */
void* ff_memalloc (word msize);   /* Allocate memory block */
void ff_memfree (void* mblock);   /* Free memory block */
#endif
#endif





/*--------------------------------------------------------------*/
/* Flags and offset address                                     */


/* File access control and file status flags (FIL.flag) */

#define FA_READ    0x01
#define FA_OPEN_EXISTING 0x00
#define FA_WRITE   0x02
#define FA_CREATE_NEW  0x04
#define FA_CREATE_ALWAYS 0x08
#define FA_OPEN_ALWAYS  0x10
#define FA__WRITTEN   0x20
#define FA__DIRTY   0x40


/* FAT sub type (FATFS.fs_type) */

#define FS_FAT12 1
#define FS_FAT16 2
#define FS_FAT32 3


/* File attribute bits for directory entry */

#define AM_RDO 0x01 /* Read only */
#define AM_HID 0x02 /* Hidden */
#define AM_SYS 0x04 /* System */
#define AM_VOL 0x08 /* Volume label */
#define AM_LFN 0x0F /* LFN entry */
#define AM_DIR 0x10 /* Directory */
#define AM_ARC 0x20 /* Archive */
#define AM_MASK 0x3F /* Mask of defined bits */


/* Fast seek feature */
#define CREATE_LINKMAP 0xFFFFFFFF



/*--------------------------------*/
/* Multi-byte word access macros  */

#if _WORD_ACCESS==1 /* Enable word access to the FAT structure */
#define LD_word(ptr)  (word)(*(word*)(byte*)(ptr))
#define LD_dword(ptr)  (dword)(*(dword*)(byte*)(ptr))
#define ST_word(ptr,val) *(word*)(byte*)(ptr)=(word)(val)
#define ST_dword(ptr,val) *(dword*)(byte*)(ptr)=(dword)(val)
#else     /* Use byte-by-byte access to the FAT structure */
#define LD_word(ptr)  (word)(((word)*((byte*)(ptr)+1)<<8)|(word)*(byte*)(ptr))
#define LD_dword(ptr)  (dword)(((dword)*((byte*)(ptr)+3)<<24)|((dword)*((byte*)(ptr)+2)<<16)|((word)*((byte*)(ptr)+1)<<8)|*(byte*)(ptr))
#define ST_word(ptr,val) *(byte*)(ptr)=(byte)(val); *((byte*)(ptr)+1)=(byte)((word)(val)>>8)
#define ST_dword(ptr,val) *(byte*)(ptr)=(byte)(val); *((byte*)(ptr)+1)=(byte)((word)(val)>>8); *((byte*)(ptr)+2)=(byte)((dword)(val)>>16); *((byte*)(ptr)+3)=(byte)((dword)(val)>>24)
#endif




//Definitions on sector size
#define SS(fs) 512U   //Fixed sector size


//Reentrancy related
#define LEAVE_FF(fs, res) return res

#define ABORT(fs, res)  { fp->err=(byte)(res); LEAVE_FF(fs, res); }



/* DBCS code ranges and SBCS extend character conversion table */

#if _CODE_PAGE==866 /* Russian (OEM) */
#define _EXCVT {0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F, \
    0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF, \
    0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
    0x90,0x91,0x92,0x93,0x9d,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,0xF0,0xF0,0xF2,0xF2,0xF4,0xF4,0xF6,0xF6,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF}

#elif _CODE_PAGE==1251 /* Cyrillic (Windows) */
#define _EXCVT {0x80,0x81,0x82,0x82,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x80,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x8A,0x9B,0x8C,0x8D,0x8E,0x8F, \
    0xA0,0xA2,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB2,0xA5,0xB5,0xB6,0xB7,0xA8,0xB9,0xAA,0xBB,0xA3,0xBD,0xBD,0xAF, \
    0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF, \
    0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF}

#elif _CODE_PAGE==1 /* ASCII (for only non-LFN cfg) */

#if _USE_LFN
#error Cannot use LFN feature without valid code page.
#endif

#else
#error Unknown code page
#endif


/* Character code support macros */
#define IsUpper(c) (((c)>='A')&&((c)<='Z'))
#define IsLower(c) (((c)>='a')&&((c)<='z'))
#define IsDigit(c) (((c)>='0')&&((c)<='9'))

#define IsDBCS1(c) 0
#define IsDBCS2(c) 0


/* Name status flags */
#define NS   11  /* Index of name status byte in fn[] */
#define NS_LOSS  0x01 /* Out of 8.3 format */
#define NS_LFN  0x02 /* Force to create LFN entry */
#define NS_LAST  0x04 /* Last segment */
#define NS_BODY  0x08 /* Lower case flag (body) */
#define NS_EXT  0x10 /* Lower case flag (ext) */
#define NS_DOT  0x20 /* Dot entry */


/* FAT sub-type boundaries */
#define MIN_FAT16 4086U /* Minimum number of clusters for FAT16 */
#define MIN_FAT32 65526U /* Minimum number of clusters for FAT32 */


/* FatFs refers the members in the FAT structures as byte array instead of
/ structure member because the structure is not binary compatible between
/ different platforms */

#define BS_jmpBoot   0 /* Jump instruction (3) */
#define BS_OEMName   3 /* OEM name (8) */
#define BPB_BytsPerSec  11 /* Sector size [byte] (2) */
#define BPB_SecPerClus  13 /* Cluster size [sector] (1) */
#define BPB_RsvdSecCnt  14 /* Size of reserved area [sector] (2) */
#define BPB_NumFATs   16 /* Number of FAT copies (1) */
#define BPB_RootEntCnt  17 /* Number of root directory entries for FAT12/16 (2) */
#define BPB_TotSec16  19 /* Volume size [sector] (2) */
#define BPB_Media   21 /* Media descriptor (1) */
#define BPB_FATSz16   22 /* FAT size [sector] (2) */
#define BPB_SecPerTrk  24 /* Track size [sector] (2) */
#define BPB_NumHeads  26 /* Number of heads (2) */
#define BPB_HiddSec   28 /* Number of special hidden sectors (4) */
#define BPB_TotSec32  32 /* Volume size [sector] (4) */
#define BS_DrvNum   36 /* Physical drive number (2) */
#define BS_BootSig   38 /* Extended boot signature (1) */
#define BS_VolID   39 /* Volume serial number (4) */
#define BS_VolLab   43 /* Volume label (8) */
#define BS_FilSysType  54 /* File system type (1) */
#define BPB_FATSz32   36 /* FAT size [sector] (4) */
#define BPB_ExtFlags  40 /* Extended flags (2) */
#define BPB_FSVer   42 /* File system version (2) */
#define BPB_RootClus  44 /* Root directory first cluster (4) */
#define BPB_FSInfo   48 /* Offset of FSINFO sector (2) */
#define BPB_BkBootSec  50 /* Offset of backup boot sector (2) */
#define BS_DrvNum32   64 /* Physical drive number (2) */
#define BS_BootSig32  66 /* Extended boot signature (1) */
#define BS_VolID32   67 /* Volume serial number (4) */
#define BS_VolLab32   71 /* Volume label (8) */
#define BS_FilSysType32  82 /* File system type (1) */
#define FSI_LeadSig   0 /* FSI: Leading signature (4) */
#define FSI_StrucSig  484 /* FSI: Structure signature (4) */
#define FSI_Free_Count  488 /* FSI: Number of free clusters (4) */
#define FSI_Nxt_Free  492 /* FSI: Last allocated cluster (4) */
#define MBR_Table   446 /* MBR: Partition table offset (2) */
#define SZ_PTE    16 /* MBR: Size of a partition table entry */
#define BS_55AA    510 /* Boot sector signature (2) */

#define DIR_Name   0 /* Short file name (11) */
#define DIR_Attr   11 /* Attribute (1) */
#define DIR_NTres   12 /* NT flag (1) */
#define DIR_CrtTimeTenth 13 /* Created time sub-second (1) */
#define DIR_CrtTime   14 /* Created time (2) */
#define DIR_CrtDate   16 /* Created date (2) */
#define DIR_LstAccDate  18 /* Last accessed date (2) */
#define DIR_FstClusHI  20 /* Higher 16-bit of first cluster (2) */
#define DIR_WrtTime   22 /* Modified time (2) */
#define DIR_WrtDate   24 /* Modified date (2) */
#define DIR_FstClusLO  26 /* Lower 16-bit of first cluster (2) */
#define DIR_FileSize  28 /* File size (4) */
#define LDIR_Ord   0 /* LFN entry order and LLE flag (1) */
#define LDIR_Attr   11 /* LFN attribute (1) */
#define LDIR_Type   12 /* LFN type (1) */
#define LDIR_Chksum   13 /* Sum of corresponding SFN entry */
#define LDIR_FstClusLO  26 /* Filled by zero (0) */
#define SZ_DIR    32  /* Size of a directory entry */
#define LLE     0x40 /* Last long entry flag in LDIR_Ord */
#define DDE     0xE5 /* Deleted directory entry mark in DIR_Name[0] */
#define NDDE    0x05 /* Replacement of the character collides with DDE */


/*------------------------------------------------------------*/
/* Module private work area                                   */
/*------------------------------------------------------------*/
/* Note that uninitialized variables with static duration are
/  zeroed/nulled at start-up. If not, the compiler or start-up
/  routine is out of ANSI-C standard.
*/

static FATFS *FatFs; /* Pointer to the file system objects (logical drives) */

static
word Fsid;    /* File system mount ID */


#if _USE_LFN==0   /* No LFN feature */
#define DEF_NAMEBUF   byte sfn[12]
#define INIT_BUF(dobj)  (dobj).fn=sfn
#define FREE_BUF()

#elif _USE_LFN==1   /* LFN feature with static working buffer */
static
word LfnBuf[_MAX_LFN+1];
#define DEF_NAMEBUF   byte sfn[12]
#define INIT_BUF(dobj)  { (dobj).fn=sfn; (dobj).lfn=LfnBuf; }
#define FREE_BUF()

#elif _USE_LFN==2   /* LFN feature with dynamic working buffer on the stack */
#define DEF_NAMEBUF   byte sfn[12]; word lbuf[_MAX_LFN+1]
#define INIT_BUF(dobj)  { (dobj).fn=sfn; (dobj).lfn=lbuf; }
#define FREE_BUF()

#elif _USE_LFN==3   /* LFN feature with dynamic working buffer on the heap */
#define DEF_NAMEBUF   byte sfn[12]; word *lfn
#define INIT_BUF(dobj)  { lfn=ff_memalloc((_MAX_LFN + 1) * 2); \
         if (!lfn) LEAVE_FF((dobj).fs, FR_NOT_ENOUGH_CORE); \
         (dobj).lfn=lfn; (dobj).fn=sfn; }
#define FREE_BUF()   ff_memfree(lfn)

#else
#error Wrong LFN configuration.
#endif


#ifdef _EXCVT
static
const byte ExCvt[]=_EXCVT; /* Upper conversion table for extended characters */
#endif






/*--------------------------------------------------------------------------

   Module Private Functions

---------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/* String functions                                                      */
/*-----------------------------------------------------------------------*/

//Copy memory to memory
static
void mem_cpy (void* dst, const void* src, word cnt)
{
 byte *d=(byte*)dst;
 const byte *s=(const byte*)src;

#if _WORD_ACCESS==1
 while (cnt >= sizeof (int)){
  *(int*)d=*(int*)s;
  d += sizeof (int); s += sizeof (int);
  cnt -= sizeof (int);
 }
#endif
 while (cnt--)*d++=*s++;
}

//Fill memory
static
void mem_set (void* dst, int val, word cnt)
{
 byte *d=(byte*)dst;
 while (cnt--)*d++=(byte)val;
}

//Compare memory to memory
static
int mem_cmp (const void* dst, const void* src, word cnt)
{
 const byte *d=(const byte *)dst, *s=(const byte *)src;
 int r=0;

 while (cnt-- && (r=*d++ - *s++)==0) ;
 return r;
}

//Check if chr is contained in the string
static
int chk_chr (const char* str, int chr)
{
 while (*str && *str != chr) str++;
 return *str;
}





/*-----------------------------------------------------------------------*/
/* Move/Flush disk access window in the file system object               */
/*-----------------------------------------------------------------------*/
static FRESULT sync_window(FATFS* fs)
{
 dword wsect;
 word nf;


 if (fs->wflag) { /* Write back the sector if it is dirty */
  wsect=fs->winsect; /* Current sector number */
  if (disk_write(fs->win, wsect, 1))
   return FR_DISK_ERR;
  fs->wflag=0;
  if (wsect - fs->fatbase < fs->fsize) {  /* Is it in the FAT area? */
   for (nf=fs->n_fats; nf >= 2; nf--) { /* Reflect the change to all FAT copies */
    wsect += fs->fsize;
    disk_write(fs->win, wsect, 1);
   }
  }
 }
 return FR_OK;
}


static
FRESULT move_window (
 FATFS* fs,  /* File system object */
 dword sector /* Sector number to make appearance in the fs->win[] */
)
{
 if (sector != fs->winsect) { /* Changed current window */
  if (sync_window(fs) != FR_OK)
   return FR_DISK_ERR;
  if (disk_read(fs->win, sector, 1))
   return FR_DISK_ERR;
  fs->winsect=sector;
 }

 return FR_OK;
}




/*-----------------------------------------------------------------------*/
/* Synchronize file system and strage device                             */
/*-----------------------------------------------------------------------*/
static
FRESULT sync_fs ( /* FR_OK: successful, FR_DISK_ERR: failed */
 FATFS* fs  /* File system object */
)
{
 FRESULT res;


 res=sync_window(fs);
 if (res==FR_OK) {
  /* Update FSINFO sector if needed */
  if (fs->fs_type==FS_FAT32 && fs->fsi_flag==1) {
   /* Create FSINFO structure */
   mem_set(fs->win, 0, SS(fs));
   ST_word(fs->win+BS_55AA, 0xAA55);
   ST_dword(fs->win+FSI_LeadSig, 0x41615252);
   ST_dword(fs->win+FSI_StrucSig, 0x61417272);
   ST_dword(fs->win+FSI_Free_Count, fs->free_clust);
   ST_dword(fs->win+FSI_Nxt_Free, fs->last_clust);
   /* Write it into the FSINFO sector */
   fs->winsect=fs->volbase + 1;
   disk_write(fs->win, fs->winsect, 1);
   fs->fsi_flag=0;
  }
  /* Make sure that no pending write process in the physical drive */
  if (disk_ioctl(CTRL_SYNC, 0) != RES_OK)
   res=FR_DISK_ERR;
 }

 return res;
}




/*-----------------------------------------------------------------------*/
/* Get sector# from cluster#                                             */
/*-----------------------------------------------------------------------*/


dword clust2sect ( /* !=0: Sector number, 0: Failed - invalid cluster# */
 FATFS* fs,  /* File system object */
 dword clst  /* Cluster# to be converted */
)
{
 clst -= 2;
 if (clst >= (fs->n_fatent - 2)) return 0;  /* Invalid cluster# */
 return clst * fs->csize + fs->database;
}




/*-----------------------------------------------------------------------*/
/* FAT access - Read value of a FAT entry                                */
/*-----------------------------------------------------------------------*/


dword get_fat ( /* 0xFFFFFFFF:Disk error, 1:internal error, Else:Cluster status */
 FATFS* fs, /* File system object */
 dword clst /* Cluster# to get the link information */
)
{
 word wc, bc;
 byte *p;


 if (clst < 2 || clst >= fs->n_fatent) /* Check range */
  return 1;

 switch (fs->fs_type) {
 case FS_FAT12 :
  bc=(word)clst; bc += bc / 2;
  if (move_window(fs, fs->fatbase + (bc / SS(fs)))) break;
  wc=fs->win[bc % SS(fs)]; bc++;
  if (move_window(fs, fs->fatbase + (bc / SS(fs)))) break;
  wc |= fs->win[bc % SS(fs)] << 8;
  return clst & 1 ? wc >> 4 : (wc & 0xFFF);

 case FS_FAT16 :
  if (move_window(fs, fs->fatbase + (clst / (SS(fs) / 2)))) break;
  p=&fs->win[clst * 2 % SS(fs)];
  return LD_word(p);

 case FS_FAT32 :
  if (move_window(fs, fs->fatbase + (clst / (SS(fs) / 4)))) break;
  p=&fs->win[clst * 4 % SS(fs)];
  return LD_dword(p) & 0x0FFFFFFF;
 }

 return 0xFFFFFFFF; /* An error occurred at the disk I/O layer */
}




/*-----------------------------------------------------------------------*/
/* FAT access - Change value of a FAT entry                              */
/*-----------------------------------------------------------------------*/
FRESULT put_fat (
 FATFS* fs, /* File system object */
 dword clst, /* Cluster# to be changed in range of 2 to fs->n_fatent - 1 */
 dword val /* New value to mark the cluster */
)
{
 word bc;
 byte *p;
 FRESULT res;


 if (clst < 2 || clst >= fs->n_fatent) { /* Check range */
  res=FR_int_ERR;

 } else {
  switch (fs->fs_type) {
  case FS_FAT12 :
   bc=(word)clst; bc += bc / 2;
   res=move_window(fs, fs->fatbase + (bc / SS(fs)));
   if (res != FR_OK) break;
   p=&fs->win[bc % SS(fs)];
   *p=(clst & 1) ? ((*p & 0x0F) | ((byte)val << 4)) : (byte)val;
   bc++;
   fs->wflag=1;
   res=move_window(fs, fs->fatbase + (bc / SS(fs)));
   if (res != FR_OK) break;
   p=&fs->win[bc % SS(fs)];
   *p=(clst & 1) ? (byte)(val >> 4) : ((*p & 0xF0) | ((byte)(val >> 8) & 0x0F));
   break;

  case FS_FAT16 :
   res=move_window(fs, fs->fatbase + (clst / (SS(fs) / 2)));
   if (res != FR_OK) break;
   p=&fs->win[clst * 2 % SS(fs)];
   ST_word(p, (word)val);
   break;

  case FS_FAT32 :
   res=move_window(fs, fs->fatbase + (clst / (SS(fs) / 4)));
   if (res != FR_OK) break;
   p=&fs->win[clst * 4 % SS(fs)];
   val |= LD_dword(p) & 0xF0000000;
   ST_dword(p, val);
   break;

  default :
   res=FR_int_ERR;
  }
  fs->wflag=1;
 }

 return res;
}




/*-----------------------------------------------------------------------*/
/* FAT handling - Remove a cluster chain                                 */
/*-----------------------------------------------------------------------*/
static
FRESULT remove_chain (
 FATFS* fs,   /* File system object */
 dword clst   /* Cluster# to remove a chain from */
)
{
 FRESULT res;
 dword nxt;

 if (clst < 2 || clst >= fs->n_fatent) { /* Check range */
  res=FR_int_ERR;

 } else {
  res=FR_OK;
  while (clst < fs->n_fatent) {   /* Not a last link? */
   nxt=get_fat(fs, clst);   /* Get cluster status */
   if (nxt==0) break;    /* Empty cluster? */
   if (nxt==1) { res=FR_int_ERR; break; } /* internal error? */
   if (nxt==0xFFFFFFFF) { res=FR_DISK_ERR; break; } /* Disk error? */
   res=put_fat(fs, clst, 0);   /* Mark the cluster "empty" */
   if (res != FR_OK) break;
   if (fs->free_clust != 0xFFFFFFFF) { /* Update FSINFO */
    fs->free_clust++;
    fs->fsi_flag |= 1;
   }
   clst=nxt; /* Next cluster */
  }
 }

 return res;
}




/*-----------------------------------------------------------------------*/
/* FAT handling - Stretch or Create a cluster chain                      */
/*-----------------------------------------------------------------------*/
static
dword create_chain ( /* 0:No free cluster, 1:internal error, 0xFFFFFFFF:Disk error, >=2:New cluster# */
 FATFS* fs,   /* File system object */
 dword clst   /* Cluster# to stretch. 0 means create a new chain. */
)
{
 dword cs, ncl, scl;
 FRESULT res;


 if (clst==0) {  /* Create a new chain */
  scl=fs->last_clust;   /* Get suggested start point */
  if (!scl || scl >= fs->n_fatent) scl=1;
 }
 else {     /* Stretch the current chain */
  cs=get_fat(fs, clst);   /* Check the cluster status */
  if (cs < 2) return 1;   /* It is an invalid cluster */
  if (cs < fs->n_fatent) return cs; /* It is already followed by next cluster */
  scl=clst;
 }

 ncl=scl;    /* Start cluster */
 for (;;) {
  ncl++;       /* Next cluster */
  if (ncl >= fs->n_fatent) {  /* Wrap around */
   ncl=2;
   if (ncl > scl) return 0; /* No free cluster */
  }
  cs=get_fat(fs, ncl);   /* Get the cluster status */
  if (cs==0) break;    /* Found a free cluster */
  if (cs==0xFFFFFFFF || cs==1)/* An error occurred */
   return cs;
  if (ncl==scl) return 0;  /* No free cluster */
 }

 res=put_fat(fs, ncl, 0x0FFFFFFF); /* Mark the new cluster "last link" */
 if (res==FR_OK && clst != 0) {
  res=put_fat(fs, clst, ncl); /* Link it to the previous one if needed */
 }
 if (res==FR_OK) {
  fs->last_clust=ncl;   /* Update FSINFO */
  if (fs->free_clust != 0xFFFFFFFF) {
   fs->free_clust--;
   fs->fsi_flag |= 1;
  }
 } else {
  ncl=(res==FR_DISK_ERR) ? 0xFFFFFFFF : 1;
 }

 return ncl;  /* Return new cluster number or error code */
}




/*-----------------------------------------------------------------------*/
/* FAT handling - Convert offset into cluster with link map table        */
/*-----------------------------------------------------------------------*/

#if _USE_FASTSEEK
static
dword clmt_clust ( /* <2:Error, >=2:Cluster number */
 FIL* fp,  /* Pointer to the file object */
 dword ofs  /* File offset to be converted to cluster# */
)
{
 dword cl, ncl, *tbl;


 tbl=fp->cltbl + 1; /* Top of CLMT */
 cl=ofs / SS(fp->fs) / fp->fs->csize; /* Cluster order from top of the file */
 for (;;) {
  ncl=*tbl++;   /* Number of cluters in the fragment */
  if (!ncl) return 0;  /* End of table? (error) */
  if (cl < ncl) break; /* In this fragment? */
  cl -= ncl; tbl++;  /* Next fragment */
 }
 return cl + *tbl; /* Return the cluster number */
}
#endif /* _USE_FASTSEEK */




/*-----------------------------------------------------------------------*/
/* Directory handling - Set directory index                              */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_sdi (
 DIR* dp,  /* Pointer to directory object */
 word idx  /* Index of directory table */
)
{
 dword clst;
 word ic;


 dp->index=idx;
 clst=dp->sclust;
 if (clst==1 || clst >= dp->fs->n_fatent) /* Check start cluster range */
  return FR_int_ERR;
 if (!clst && dp->fs->fs_type==FS_FAT32) /* Replace cluster# 0 with root cluster# if in FAT32 */
  clst=dp->fs->dirbase;

 if (clst==0) { /* Static table (root-directory in FAT12/16) */
  dp->clust=clst;
  if (idx >= dp->fs->n_rootdir)  /* Index is out of range */
   return FR_int_ERR;
  dp->sect=dp->fs->dirbase + idx / (SS(dp->fs) / SZ_DIR); /* Sector# */
 }
 else {    /* Dynamic table (sub-dirs or root-directory in FAT32) */
  ic=SS(dp->fs) / SZ_DIR * dp->fs->csize; /* Entries per cluster */
  while (idx >= ic) { /* Follow cluster chain */
   clst=get_fat(dp->fs, clst);    /* Get next cluster */
   if (clst==0xFFFFFFFF) return FR_DISK_ERR; /* Disk error */
   if (clst < 2 || clst >= dp->fs->n_fatent) /* Reached to end of table or int error */
    return FR_int_ERR;
   idx -= ic;
  }
  dp->clust=clst;
  dp->sect=clust2sect(dp->fs, clst) + idx / (SS(dp->fs) / SZ_DIR); /* Sector# */
 }

 dp->dir=dp->fs->win + (idx % (SS(dp->fs) / SZ_DIR)) * SZ_DIR; /* Ptr to the entry in the sector */

 return FR_OK; /* Seek succeeded */
}




/*-----------------------------------------------------------------------*/
/* Directory handling - Move directory table index next                  */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_next ( /* FR_OK:Succeeded, FR_NO_FILE:End of table, FR_DENIED:Could not stretch */
 DIR* dp,  /* Pointer to the directory object */
 int stretch  /* 0: Do not stretch table, 1: Stretch table if needed */
)
{
 dword clst;
 word i;


 i=dp->index + 1;
 if (!i || !dp->sect) /* Report EOT when index has reached 65535 */
  return FR_NO_FILE;

 if (!(i % (SS(dp->fs) / SZ_DIR))) { /* Sector changed? */
  dp->sect++;     /* Next sector */

  if (!dp->clust) {  /* Static table */
   if (i >= dp->fs->n_rootdir) /* Report EOT if it reached end of static table */
    return FR_NO_FILE;
  }
  else {     /* Dynamic table */
   if (((i / (SS(dp->fs) / SZ_DIR)) & (dp->fs->csize - 1))==0) { /* Cluster changed? */
    clst=get_fat(dp->fs, dp->clust);    /* Get next cluster */
    if (clst <= 1) return FR_int_ERR;
    if (clst==0xFFFFFFFF) return FR_DISK_ERR;
    if (clst >= dp->fs->n_fatent) {     /* If it reached end of dynamic table, */
     byte c;
     if (!stretch) return FR_NO_FILE;   /* If do not stretch, report EOT */
     clst=create_chain(dp->fs, dp->clust);  /* Stretch cluster chain */
     if (clst==0) return FR_DENIED;   /* No free cluster */
     if (clst==1) return FR_int_ERR;
     if (clst==0xFFFFFFFF) return FR_DISK_ERR;
     /* Clean-up stretched table */
     if (sync_window(dp->fs)) return FR_DISK_ERR;/* Flush disk access window */
     mem_set(dp->fs->win, 0, SS(dp->fs));  /* Clear window buffer */
     dp->fs->winsect=clust2sect(dp->fs, clst); /* Cluster start sector */
     for (c=0; c < dp->fs->csize; c++) {  /* Fill the new cluster with 0 */
      dp->fs->wflag=1;
      if (sync_window(dp->fs)) return FR_DISK_ERR;
      dp->fs->winsect++;
     }
     dp->fs->winsect -= c;      /* Rewind window offset */
    }
    dp->clust=clst;    /* Initialize data for new cluster */
    dp->sect=clust2sect(dp->fs, clst);
   }
  }
 }

 dp->index=i; /* Current index */
 dp->dir=dp->fs->win + (i % (SS(dp->fs) / SZ_DIR)) * SZ_DIR; /* Current entry in the window */

 return FR_OK;
}




/*-----------------------------------------------------------------------*/
/* Directory handling - Reserve directory entry                          */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_alloc (
 DIR* dp, /* Pointer to the directory object */
 word nent /* Number of contiguous entries to allocate (1-21) */
)
{
 FRESULT res;
 word n;


 res=dir_sdi(dp, 0);
 if (res==FR_OK) {
  n=0;
  do {
   res=move_window(dp->fs, dp->sect);
   if (res != FR_OK) break;
   if (dp->dir[0]==DDE || dp->dir[0]==0) { /* Is it a blank entry? */
    if (++n==nent) break; /* A block of contiguous entries is found */
   } else {
    n=0;     /* Not a blank entry. Restart to search */
   }
   res=dir_next(dp, 1);  /* Next entry with table stretch enabled */
  } while (res==FR_OK);
 }
 if (res==FR_NO_FILE) res=FR_DENIED; /* No directory entry to allocate */
 return res;
}




/*-----------------------------------------------------------------------*/
/* Directory handling - Load/Store start cluster number                  */
/*-----------------------------------------------------------------------*/

static
dword ld_clust (
 FATFS* fs, /* Pointer to the fs object */
 byte* dir /* Pointer to the directory entry */
)
{
 dword cl;

 cl=LD_word(dir+DIR_FstClusLO);
 if (fs->fs_type==FS_FAT32)
  cl |= (dword)LD_word(dir+DIR_FstClusHI) << 16;

 return cl;
}


static
void st_clust (
 byte* dir, /* Pointer to the directory entry */
 dword cl /* Value to be set */
)
{
 ST_word(dir+DIR_FstClusLO, cl);
 ST_word(dir+DIR_FstClusHI, cl >> 16);
}




/*-----------------------------------------------------------------------*/
/* LFN handling - Test/Pick/Fit an LFN segment from/to directory entry   */
/*-----------------------------------------------------------------------*/
#if _USE_LFN
static
const byte LfnOfs[]={1,3,5,7,9,14,16,18,20,22,24,28,30}; /* Offset of LFN characters in the directory entry */


static
int cmp_lfn (   /* 1:Matched, 0:Not matched */
 word* lfnbuf,  /* Pointer to the LFN to be compared */
 byte* dir   /* Pointer to the directory entry containing a part of LFN */
)
{
 word i, s;
 word wc, uc;


 i=((dir[LDIR_Ord] & ~LLE) - 1) * 13; /* Get offset in the LFN buffer */
 s=0; wc=1;
 do {
  uc=LD_word(dir+LfnOfs[s]); /* Pick an LFN character from the entry */
  if (wc) { /* Last character has not been processed */
   wc=ff_wtoupper(uc);  /* Convert it to upper case */
   if (i >= _MAX_LFN || wc != ff_wtoupper(lfnbuf[i++])) /* Compare it */
    return 0;    /* Not matched */
  } else {
   if (uc != 0xFFFF) return 0; /* Check filler */
  }
 } while (++s < 13);    /* Repeat until all characters in the entry are checked */

 if ((dir[LDIR_Ord] & LLE) && wc && lfnbuf[i]) /* Last segment matched but different length */
  return 0;

 return 1;      /* The part of LFN matched */
}



static
int pick_lfn (   /* 1:Succeeded, 0:Buffer overflow */
 word* lfnbuf,  /* Pointer to the Unicode-LFN buffer */
 byte* dir   /* Pointer to the directory entry */
)
{
 word i, s;
 word wc, uc;


 i=((dir[LDIR_Ord] & 0x3F) - 1) * 13; /* Offset in the LFN buffer */

 s=0; wc=1;
 do {
  uc=LD_word(dir+LfnOfs[s]);  /* Pick an LFN character from the entry */
  if (wc) { /* Last character has not been processed */
   if (i >= _MAX_LFN) return 0; /* Buffer overflow? */
   lfnbuf[i++]=wc=uc;   /* Store it */
  } else {
   if (uc != 0xFFFF) return 0;  /* Check filler */
  }
 } while (++s < 13);      /* Read all character in the entry */

 if (dir[LDIR_Ord] & LLE) {    /* Put terminator if it is the last LFN part */
  if (i >= _MAX_LFN) return 0;  /* Buffer overflow? */
  lfnbuf[i]=0;
 }

 return 1;
}


static
void fit_lfn (
 const word* lfnbuf, /* Pointer to the LFN buffer */
 byte* dir,    /* Pointer to the directory entry */
 byte ord,    /* LFN order (1-20) */
 byte sum    /* SFN sum */
)
{
 word i, s;
 word wc;


 dir[LDIR_Chksum]=sum;   /* Set check sum */
 dir[LDIR_Attr]=AM_LFN;  /* Set attribute. LFN entry */
 dir[LDIR_Type]=0;
 ST_word(dir+LDIR_FstClusLO, 0);

 i=(ord - 1) * 13;    /* Get offset in the LFN buffer */
 s=wc=0;
 do {
  if (wc != 0xFFFF) wc=lfnbuf[i++]; /* Get an effective character */
  ST_word(dir+LfnOfs[s], wc); /* Put it */
  if (!wc) wc=0xFFFF;  /* Padding characters following last character */
 } while (++s < 13);
 if (wc==0xFFFF || !lfnbuf[i]) ord |= LLE; /* Bottom LFN part is the start of LFN sequence */
 dir[LDIR_Ord]=ord;   /* Set the LFN order */
}

#endif




/*-----------------------------------------------------------------------*/
/* Create numbered name                                                  */
/*-----------------------------------------------------------------------*/
#if _USE_LFN
void gen_numname (
 byte* dst,   /* Pointer to generated SFN */
 const byte* src, /* Pointer to source SFN to be modified */
 const word* lfn, /* Pointer to LFN */
 word seq   /* Sequence number */
)
{
 byte ns[8], c;
 word i, j;


 mem_cpy(dst, src, 11);

 if (seq > 5) { /* On many collisions, generate a hash number instead of sequential number */
  do seq=(seq >> 1) + (seq << 15) + (word)*lfn++; while (*lfn);
 }

 /* itoa (hexdecimal) */
 i=7;
 do {
  c=(seq % 16) + '0';
  if (c > '9') c += 7;
  ns[i--]=c;
  seq /= 16;
 } while (seq);
 ns[i]='~';

 /* Append the number */
 for (j=0; j < i && dst[j] != ' '; j++) {
  if (IsDBCS1(dst[j])) {
   if (j==i - 1) break;
   j++;
  }
 }
 do {
  dst[j++]=(i < 8) ? ns[i++] : ' ';
 } while (j < 8);
}
#endif




/*-----------------------------------------------------------------------*/
/* Calculate sum of an SFN                                               */
/*-----------------------------------------------------------------------*/
#if _USE_LFN
static
byte sum_sfn (
 const byte* dir  /* Pointer to the SFN entry */
)
{
 byte sum=0;
 word n=11;

 do sum=(sum >> 1) + (sum << 7) + *dir++; while (--n);
 return sum;
}
#endif




/*-----------------------------------------------------------------------*/
/* Directory handling - Find an object in the directory                  */
/*-----------------------------------------------------------------------*/

static
FRESULT dir_find (
 DIR* dp   /* Pointer to the directory object linked to the file name */
)
{
 FRESULT res;
 byte c, *dir;
#if _USE_LFN
 byte a, ord, sum;
#endif

 res=dir_sdi(dp, 0);   /* Rewind directory object */
 if (res != FR_OK) return res;

#if _USE_LFN
 ord=sum=0xFF;
#endif
 do {
  res=move_window(dp->fs, dp->sect);
  if (res != FR_OK) break;
  dir=dp->dir;     /* Ptr to the directory entry of current index */
  c=dir[DIR_Name];
  if (c==0) { res=FR_NO_FILE; break; } /* Reached to end of table */
#if _USE_LFN /* LFN configuration */
  a=dir[DIR_Attr] & AM_MASK;
  if (c==DDE || ((a & AM_VOL) && a != AM_LFN)) { /* An entry without valid data */
   ord=0xFF;
  } else {
   if (a==AM_LFN) {   /* An LFN entry is found */
    if (dp->lfn) {
     if (c & LLE) {  /* Is it start of LFN sequence? */
      sum=dir[LDIR_Chksum];
      c &= ~LLE; ord=c; /* LFN start order */
      dp->lfn_idx=dp->index;
     }
     /* Check validity of the LFN entry and compare it with given name */
     ord=(c==ord && sum==dir[LDIR_Chksum] && cmp_lfn(dp->lfn, dir)) ? ord - 1 : 0xFF;
    }
   } else {     /* An SFN entry is found */
    if (!ord && sum==sum_sfn(dir)) break; /* LFN matched? */
    ord=0xFF; dp->lfn_idx=0xFFFF; /* Reset LFN sequence */
    if (!(dp->fn[NS] & NS_LOSS) && !mem_cmp(dir, dp->fn, 11)) break; /* SFN matched? */
   }
  }
#else  /* Non LFN configuration */
  if (!(dir[DIR_Attr] & AM_VOL) && !mem_cmp(dir, dp->fn, 11)) /* Is it a valid entry? */
   break;
#endif
  res=dir_next(dp, 0);  /* Next entry */
 } while (res==FR_OK);

 return res;
}




/*-----------------------------------------------------------------------*/
/* Read an object from the directory                                     */
/*-----------------------------------------------------------------------*/
#if _FS_MINIMIZE <= 1 || _USE_LABEL 
static
FRESULT dir_read (
 DIR* dp,  /* Pointer to the directory object */
 int vol   /* Filtered by 0:file/directory or 1:volume label */
)
{
 FRESULT res;
 byte a, c, *dir;
#if _USE_LFN
 byte ord=0xFF, sum=0xFF;
#endif

 res=FR_NO_FILE;
 while (dp->sect) {
  res=move_window(dp->fs, dp->sect);
  if (res != FR_OK) break;
  dir=dp->dir;     /* Ptr to the directory entry of current index */
  c=dir[DIR_Name];
  if (c==0) { res=FR_NO_FILE; break; } /* Reached to end of table */
  a=dir[DIR_Attr] & AM_MASK;
#if _USE_LFN /* LFN configuration */
  if (c==DDE || (c=='.') || (int)(a==AM_VOL) != vol) { /* An entry without valid data */
   ord=0xFF;
  } else {
   if (a==AM_LFN) {   /* An LFN entry is found */
    if (c & LLE) {   /* Is it start of LFN sequence? */
     sum=dir[LDIR_Chksum];
     c &= ~LLE; ord=c;
     dp->lfn_idx=dp->index;
    }
    /* Check LFN validity and capture it */
    ord=(c==ord && sum==dir[LDIR_Chksum] && pick_lfn(dp->lfn, dir)) ? ord - 1 : 0xFF;
   } else {     /* An SFN entry is found */
    if (ord || sum != sum_sfn(dir)) /* Is there a valid LFN? */
     dp->lfn_idx=0xFFFF;  /* It has no LFN. */
    break;
   }
  }
#else  /* Non LFN configuration */
  if (c != DDE && (c != '.') && a != AM_LFN && (int)(a==AM_VOL)==vol) /* Is it a valid entry? */
   break;
#endif
  res=dir_next(dp, 0);    /* Next entry */
  if (res != FR_OK) break;
 }

 if (res != FR_OK) dp->sect=0;

 return res;
}
#endif /* _FS_MINIMIZE <= 1 || _USE_LABEL || */




/*-----------------------------------------------------------------------*/
/* Register an object to the directory                                   */
/*-----------------------------------------------------------------------*/
static
FRESULT dir_register ( /* FR_OK:Successful, FR_DENIED:No free entry or too many SFN collision, FR_DISK_ERR:Disk error */
 DIR* dp    /* Target directory with object name to be created */
)
{
 FRESULT res;
#if _USE_LFN /* LFN configuration */
 word n, ne;
 byte sn[12], *fn, sum;
 word *lfn;


 fn=dp->fn; lfn=dp->lfn;
 mem_cpy(sn, fn, 12);

 if (sn[NS] & NS_LOSS) {   /* When LFN is out of 8.3 format, generate a numbered name */
  fn[NS]=0; dp->lfn=0;   /* Find only SFN */
  for (n=1; n < 100; n++) {
   gen_numname(fn, sn, lfn, n); /* Generate a numbered name */
   res=dir_find(dp);    /* Check if the name collides with existing SFN */
   if (res != FR_OK) break;
  }
  if (n==100) return FR_DENIED;  /* Abort if too many collisions */
  if (res != FR_NO_FILE) return res; /* Abort if the result is other than 'not collided' */
  fn[NS]=sn[NS]; dp->lfn=lfn;
 }

 if (sn[NS] & NS_LFN) {   /* When LFN is to be created, allocate entries for an SFN + LFNs. */
  for (n=0; lfn[n]; n++) ;
  ne=(n + 25) / 13;
 } else {      /* Otherwise allocate an entry for an SFN  */
  ne=1;
 }
 res=dir_alloc(dp, ne);  /* Allocate entries */

 if (res==FR_OK && --ne) {  /* Set LFN entry if needed */
  res=dir_sdi(dp, (word)(dp->index - ne));
  if (res==FR_OK) {
   sum=sum_sfn(dp->fn); /* Sum value of the SFN tied to the LFN */
   do {     /* Store LFN entries in bottom first */
    res=move_window(dp->fs, dp->sect);
    if (res != FR_OK) break;
    fit_lfn(dp->lfn, dp->dir, (byte)ne, sum);
    dp->fs->wflag=1;
    res=dir_next(dp, 0); /* Next entry */
   } while (res==FR_OK && --ne);
  }
 }
#else /* Non LFN configuration */
 res=dir_alloc(dp, 1);  /* Allocate an entry for SFN */
#endif

 if (res==FR_OK) {    /* Set SFN entry */
  res=move_window(dp->fs, dp->sect);
  if (res==FR_OK) {
   mem_set(dp->dir, 0, SZ_DIR); /* Clean the entry */
   mem_cpy(dp->dir, dp->fn, 11); /* Put SFN */
#if _USE_LFN
   dp->dir[DIR_NTres]=dp->fn[NS] & (NS_BODY | NS_EXT); /* Put NT flag */
#endif
   dp->fs->wflag=1;
  }
 }

 return res;
}




/*-----------------------------------------------------------------------*/
/* Remove an object from the directory                                   */
/*-----------------------------------------------------------------------*/
#if !_FS_MINIMIZE
static
FRESULT dir_remove ( /* FR_OK: Successful, FR_DISK_ERR: A disk error */
 DIR* dp    /* Directory object pointing the entry to be removed */
)
{
 FRESULT res;
#if _USE_LFN /* LFN configuration */
 word i;

 i=dp->index; /* SFN index */
 res=dir_sdi(dp, (word)((dp->lfn_idx==0xFFFF) ? i : dp->lfn_idx)); /* Goto the SFN or top of the LFN entries */
 if (res==FR_OK) {
  do {
   res=move_window(dp->fs, dp->sect);
   if (res != FR_OK) break;
   *dp->dir=DDE;   /* Mark the entry "deleted" */
   dp->fs->wflag=1;
   if (dp->index >= i) break; /* When reached SFN, all entries of the object has been deleted. */
   res=dir_next(dp, 0);  /* Next entry */
  } while (res==FR_OK);
  if (res==FR_NO_FILE) res=FR_int_ERR;
 }

#else   /* Non LFN configuration */
 res=dir_sdi(dp, dp->index);
 if (res==FR_OK) {
  res=move_window(dp->fs, dp->sect);
  if (res==FR_OK) {
   *dp->dir=DDE;   /* Mark the entry "deleted" */
   dp->fs->wflag=1;
  }
 }
#endif

 return res;
}
#endif




/*-----------------------------------------------------------------------*/
/* Pick a segment and create the object name in directory form           */
/*-----------------------------------------------------------------------*/

static
FRESULT create_name (
 DIR* dp,   /* Pointer to the directory object */
 const char** path /* Pointer to pointer to the segment in the path string */
)
{
#if _USE_LFN /* LFN configuration */
 byte b, cf;
 word w, *lfn;
 word i, ni, si, di;
 const char *p;

 /* Create LFN in Unicode */
 for (p=*path; *p=='/' || *p=='\\'; p++) ; /* Strip duplicated separator */
 lfn=dp->lfn;
 si=di=0;
 for (;;) {
  w=p[si++];     /* Get a character */
  if (w < ' ' || w=='/' || w=='\\') break; /* Break on end of segment */
  if (di >= _MAX_LFN)    /* Reject too long name */
   return FR_INVALID_NAME;
  w &= 0xFF;
  if (IsDBCS1(w)) {    /* Check if it is a DBC 1st byte (always false on SBCS cfg) */
   b=(byte)p[si++];   /* Get 2nd byte */
   if (!IsDBCS2(b))
    return FR_INVALID_NAME; /* Reject invalid sequence */
   w=(w << 8) + b;   /* Create a DBC */
  }
  w=ff_convert(w, 1);   /* Convert ANSI/OEM to Unicode */
  if (!w) return FR_INVALID_NAME; /* Reject invalid code */
  if (w < 0x80 && chk_chr("\"*:<>\?|\x7F", w)) /* Reject illegal characters for LFN */
   return FR_INVALID_NAME;
  lfn[di++]=w;     /* Store the Unicode character */
 }
 *path=&p[si];      /* Return pointer to the next segment */
 cf=(w < ' ') ? NS_LAST : 0;  /* Set last segment flag if end of path */

 while (di) {      /* Strip trailing spaces and dots */
  w=lfn[di-1];
  if (w != ' ' && w != '.') break;
  di--;
 }
 if (!di) return FR_INVALID_NAME; /* Reject nul string */

 lfn[di]=0;      /* LFN is created */

 /* Create SFN in directory form */
 mem_set(dp->fn, ' ', 11);
 for (si=0; lfn[si]==' ' || lfn[si]=='.'; si++) ; /* Strip leading spaces and dots */
 if (si) cf |= NS_LOSS | NS_LFN;
 while (di && lfn[di - 1] != '.') di--; /* Find extension (di<=si: no extension) */

 b=i=0; ni=8;
 for (;;) {
  w=lfn[si++];     /* Get an LFN character */
  if (!w) break;     /* Break on end of the LFN */
  if (w==' ' || (w=='.' && si != di)) { /* Remove spaces and dots */
   cf |= NS_LOSS | NS_LFN; continue;
  }

  if (i >= ni || si==di) {  /* Extension or end of SFN */
   if (ni==11) {    /* Long extension */
    cf |= NS_LOSS | NS_LFN; break;
   }
   if (si != di) cf |= NS_LOSS | NS_LFN; /* Out of 8.3 format */
   if (si > di) break;   /* No extension */
   si=di; i=8; ni=11; /* Enter extension section */
   b <<= 2; continue;
  }

  if (w >= 0x80) {    /* Non ASCII character */
#ifdef _EXCVT
   w=ff_convert(w, 0);  /* Unicode -> OEM code */
   if (w) w=ExCvt[w - 0x80]; /* Convert extended character to upper (SBCS) */
#else
   w=ff_convert(ff_wtoupper(w), 0); /* Upper converted Unicode -> OEM code */
#endif
   cf |= NS_LFN;    /* Force create LFN entry */
  }


  if (!w || chk_chr("+,;=[]", w)) { /* Replace illegal characters for SFN */
   w='_'; cf |= NS_LOSS | NS_LFN;/* Lossy conversion */
  } else {
   if (IsUpper(w)) {  /* ASCII large capital */
    b |= 2;
   } else {
    if (IsLower(w)) { /* ASCII small capital */
     b |= 1; w -= 0x20;
    }
   }
  }
  dp->fn[i++]=(byte)w;
 }

 if (dp->fn[0]==DDE) dp->fn[0]=NDDE; /* If the first character collides with deleted mark, replace it with 0x05 */

 if (ni==8) b <<= 2;
 if ((b & 0x0C)==0x0C || (b & 0x03)==0x03) /* Create LFN entry when there are composite capitals */
  cf |= NS_LFN;
 if (!(cf & NS_LFN)) {      /* When LFN is in 8.3 format without extended character, NT flags are created */
  if ((b & 0x03)==0x01) cf |= NS_EXT; /* NT flag (Extension has only small capital) */
  if ((b & 0x0C)==0x04) cf |= NS_BODY; /* NT flag (Filename has only small capital) */
 }

 dp->fn[NS]=cf; /* SFN is created */

 return FR_OK;


#else /* Non-LFN configuration */
 byte b, c, d, *sfn;
 word ni, si, i;
 const char *p;

 /* Create file name in directory form */
 for (p=*path; *p=='/' || *p=='\\'; p++) ; /* Strip duplicated separator */
 sfn=dp->fn;
 mem_set(sfn, ' ', 11);
 si=i=b=0; ni=8;

 for (;;) {
  c=(byte)p[si++];
  if (c <= ' ' || c=='/' || c=='\\') break; /* Break on end of segment */
  if (c=='.' || i >= ni) {
   if (ni != 8 || c != '.') return FR_INVALID_NAME;
   i=8; ni=11;
   b <<= 2; continue;
  }
  if (c >= 0x80) {    /* Extended character? */
   b |= 3;      /* Eliminate NT flag */
#ifdef _EXCVT
   c=ExCvt[c - 0x80];  /* To upper extended characters (SBCS cfg) */
#else
   return FR_INVALID_NAME;  /* Reject extended characters (ASCII cfg) */
#endif
  }
  if (IsDBCS1(c)) {    /* Check if it is a DBC 1st byte (always false on SBCS cfg) */
   d=(byte)p[si++];   /* Get 2nd byte */
   if (!IsDBCS2(d) || i >= ni - 1) /* Reject invalid DBC */
    return FR_INVALID_NAME;
   sfn[i++]=c;
   sfn[i++]=d;
  } else {      /* Single byte code */
   if (chk_chr("\"*+,:;<=>\?[]|\x7F", c)) /* Reject illegal chrs for SFN */
    return FR_INVALID_NAME;
   if (IsUpper(c)) {   /* ASCII large capital? */
    b |= 2;
   } else {
    if (IsLower(c)) {  /* ASCII small capital? */
     b |= 1; c -= 0x20;
    }
   }
   sfn[i++]=c;
  }
 }
 *path=&p[si];      /* Return pointer to the next segment */
 c=(c <= ' ') ? NS_LAST : 0;  /* Set last segment flag if end of path */

 if (!i) return FR_INVALID_NAME;  /* Reject nul string */
 if (sfn[0]==DDE) sfn[0]=NDDE; /* When first character collides with DDE, replace it with 0x05 */

 if (ni==8) b <<= 2;
 if ((b & 0x03)==0x01) c |= NS_EXT; /* NT flag (Name extension has only small capital) */
 if ((b & 0x0C)==0x04) c |= NS_BODY; /* NT flag (Name body has only small capital) */

 sfn[NS]=c;  /* Store NT flag, File name is created */

 return FR_OK;
#endif
}




/*-----------------------------------------------------------------------*/
/* Get file information from directory entry                             */
/*-----------------------------------------------------------------------*/
#if _FS_MINIMIZE <= 1
static
void get_fileinfo (  /* No return code */
 DIR* dp,   /* Pointer to the directory object */
 FILINFO* fno   /* Pointer to the file information to be filled */
)
{
 word i;
 char *p, c;


 p=fno->fname;
 if (dp->sect) {  /* Get SFN */
  byte *dir=dp->dir;

  i=0;
  while (i < 11) {  /* Copy name body and extension */
   c=(char)dir[i++];
   if (c==' ') continue;   /* Skip padding spaces */
   if (c==NDDE) c=(char)DDE; /* Restore replaced DDE character */
   if (i==9) *p++='.';   /* Insert a . if extension is exist */
#if _USE_LFN
   if (IsUpper(c) && (dir[DIR_NTres] & (i >= 9 ? NS_EXT : NS_BODY)))
    c += 0x20;   /* To lower */
#endif
   *p++=c;
  }
  fno->fattrib=dir[DIR_Attr];    /* Attribute */
  fno->fsize=LD_dword(dir+DIR_FileSize); /* Size */
  fno->fdate=LD_word(dir+DIR_WrtDate);  /* Date */
  fno->ftime=LD_word(dir+DIR_WrtTime);  /* Time */
 }
 *p=0;  /* Terminate SFN string by a \0 */

#if _USE_LFN
 if (fno->lfname) {
  word w, *lfn;

  i=0; p=fno->lfname;
  if (dp->sect && fno->lfsize && dp->lfn_idx != 0xFFFF) { /* Get LFN if available */
   lfn=dp->lfn;
   while ((w=*lfn++) != 0) {  /* Get an LFN character */
    w=ff_convert(w, 0);  /* Unicode -> OEM */
    if (!w) { i=0; break; } /* No LFN if it could not be converted */
    if (i >= fno->lfsize - 1) { i=0; break; } /* No LFN if buffer overflow */
    p[i++]=(char)w;
   }
  }
  p[i]=0; /* Terminate LFN string by a \0 */
 }
#endif
}
#endif /* _FS_MINIMIZE <= 1 */






/*-----------------------------------------------------------------------*/
/* Follow a file path                                                    */
/*-----------------------------------------------------------------------*/

static
FRESULT follow_path ( /* FR_OK(0): successful, !=0: error code */
 DIR* dp,   /* Directory object to return last directory and found object */
 const char* path /* Full-path string to find a file or directory */
)
{
 FRESULT res;
 byte *dir, ns;

 if (*path=='/' || *path=='\\')  /* Strip heading separator if exist */
  path++;
 dp->sclust=0;       /* Always start from the root directory */

 if ((word)*path < ' ') {    /* Null path name is the origin directory itself */
  res=dir_sdi(dp, 0);
  dp->dir=0;
 } else {        /* Follow path */
  for (;;) {
   res=create_name(dp, &path); /* Get a segment name of the path */
   if (res != FR_OK) break;
   res=dir_find(dp);    /* Find an object with the sagment name */
   ns=dp->fn[NS];
   if (res != FR_OK) {    /* Failed to find the object */
    if (res==FR_NO_FILE) { /* Object is not found */
     if (!(ns & NS_LAST)) res=FR_NO_PATH; /* Adjust error code if not last segment */
    }
    break;
   }
   if (ns & NS_LAST) break;   /* Last segment matched. Function completed. */
   dir=dp->dir;      /* Follow the sub-directory */
   if (!(dir[DIR_Attr] & AM_DIR)) { /* It is not a sub-directory and cannot follow */
    res=FR_NO_PATH; break;
   }
   dp->sclust=ld_clust(dp->fs, dir);
  }
 }

 return res;
}




/*-----------------------------------------------------------------------*/
/* Load a sector and check if it is an FAT boot sector                   */
/*-----------------------------------------------------------------------*/

static
byte check_fs ( /* 0:FAT boor sector, 1:Valid boor sector but not FAT, 2:Not a boot sector, 3:Disk error */
 FATFS* fs, /* File system object */
 dword sect /* Sector# (lba) to check if it is an FAT boot record or not */
)
{
 fs->wflag=0; fs->winsect=0xFFFFFFFF; /* Invaidate window */
 if (move_window(fs, sect) != FR_OK)   /* Load boot record */
  return 3;

 if (LD_word(&fs->win[BS_55AA]) != 0xAA55) /* Check boot record signature (always placed at offset 510 even if the sector size is >512) */
  return 2;

 if ((LD_dword(&fs->win[BS_FilSysType]) & 0xFFFFFF)==0x544146)  /* Check "FAT" string */
  return 0;
 if ((LD_dword(&fs->win[BS_FilSysType32]) & 0xFFFFFF)==0x544146) /* Check "FAT" string */
  return 0;

 return 1;
}




/*-----------------------------------------------------------------------*/
/* Find logical drive and check if the volume is mounted                 */
/*-----------------------------------------------------------------------*/

static
FRESULT find_volume ( /* FR_OK(0): successful, !=0: any error occurred */
 FATFS** rfs  /* Pointer to pointer to the found file system object */
)
{
 byte fmt;
 int vol;
 byte stat;
 dword bsect, fasize, tsect, sysect, nclst, szbfat;
 word nrsv;
 FATFS *fs;


 /* Get logical drive number from the path name */
 *rfs=0;
 vol=0;

 /* Check if the file system object is valid or not */
 fs=FatFs;     /* Get pointer to the file system object */
 if (!fs) return FR_NOT_ENABLED;  /* Is the file system object available? */

 *rfs=fs;       /* Return pointer to the file system object */

 if (fs->fs_type) {     /* If the volume has been mounted */
  return FR_OK;    /* The file system object is valid */
 }

 /* The file system object is not valid. */
 /* Following code attempts to mount the volume. (analyze BPB and initialize the fs object) */

 fs->fs_type=0;     /* Clear the file system object */
 stat=disk_initialize(0); /* Initialize the physical drive */
 if (stat & STA_NOINIT)    /* Check if the initialization succeeded */
  return FR_NOT_READY;   /* Failed to initialize due to no medium or hard error */
 /* Find an FAT partition on the drive. Supports only generic partitioning, FDISK and SFD. */
 bsect=0;
 fmt=check_fs(fs, bsect);     /* Load sector 0 and check if it is an FAT boot sector as SFD */
 if (fmt==1 || (!fmt && (0))) { /* Not an FAT boot sector or forced partition number */
  word i;
  dword br[4];

  for (i=0; i < 4; i++) {   /* Get partition offset */
   byte *pt=fs->win+MBR_Table + i * SZ_PTE;
   br[i]=pt[4] ? LD_dword(&pt[8]) : 0;
  }
  i=0;      /* Partition number: 0:auto, 1-4:forced */
  if (i) i--;
  do {        /* Find an FAT volume */
   bsect=br[i];
   fmt=bsect ? check_fs(fs, bsect) : 2; /* Check the partition */
  } while (!0 && fmt && ++i < 4);
 }
 if (fmt==3) return FR_DISK_ERR;  /* An error occured in the disk I/O layer */
 if (fmt) return FR_NO_FILESYSTEM;  /* No FAT volume is found */

 /* An FAT volume is found. Following code initializes the file system object */

 if (LD_word(fs->win+BPB_BytsPerSec) != SS(fs))  /* (BPB_BytsPerSec must be equal to the physical sector size) */
  return FR_NO_FILESYSTEM;

 fasize=LD_word(fs->win+BPB_FATSz16);    /* Number of sectors per FAT */
 if (!fasize) fasize=LD_dword(fs->win+BPB_FATSz32);
 fs->fsize=fasize;

 fs->n_fats=fs->win[BPB_NumFATs];     /* Number of FAT copies */
 if (fs->n_fats != 1 && fs->n_fats != 2)    /* (Must be 1 or 2) */
  return FR_NO_FILESYSTEM;
 fasize *= fs->n_fats;        /* Number of sectors for FAT area */

 fs->csize=fs->win[BPB_SecPerClus];    /* Number of sectors per cluster */
 if (!fs->csize || (fs->csize & (fs->csize - 1))) /* (Must be power of 2) */
  return FR_NO_FILESYSTEM;

 fs->n_rootdir=LD_word(fs->win+BPB_RootEntCnt); /* Number of root directory entries */
 if (fs->n_rootdir % (SS(fs) / SZ_DIR))    /* (Must be sector aligned) */
  return FR_NO_FILESYSTEM;

 tsect=LD_word(fs->win+BPB_TotSec16);    /* Number of sectors on the volume */
 if (!tsect) tsect=LD_dword(fs->win+BPB_TotSec32);

 nrsv=LD_word(fs->win+BPB_RsvdSecCnt);    /* Number of reserved sectors */
 if (!nrsv) return FR_NO_FILESYSTEM;     /* (Must not be 0) */

 /* Determine the FAT sub type */
 sysect=nrsv + fasize + fs->n_rootdir / (SS(fs) / SZ_DIR); /* RSV+FAT+DIR */
 if (tsect < sysect) return FR_NO_FILESYSTEM;  /* (Invalid volume size) */
 nclst=(tsect - sysect) / fs->csize;    /* Number of clusters */
 if (!nclst) return FR_NO_FILESYSTEM;    /* (Invalid volume size) */
 fmt=FS_FAT12;
 if (nclst >= MIN_FAT16) fmt=FS_FAT16;
 if (nclst >= MIN_FAT32) fmt=FS_FAT32;

 /* Boundaries and Limits */
 fs->n_fatent=nclst + 2;       /* Number of FAT entries */
 fs->volbase=bsect;        /* Volume start sector */
 fs->fatbase=bsect + nrsv;       /* FAT start sector */
 fs->database=bsect + sysect;      /* Data start sector */
 if (fmt==FS_FAT32) {
  if (fs->n_rootdir) return FR_NO_FILESYSTEM;  /* (BPB_RootEntCnt must be 0) */
  fs->dirbase=LD_dword(fs->win+BPB_RootClus); /* Root directory start cluster */
  szbfat=fs->n_fatent * 4;      /* (Required FAT size) */
 } else {
  if (!fs->n_rootdir) return FR_NO_FILESYSTEM; /* (BPB_RootEntCnt must not be 0) */
  fs->dirbase=fs->fatbase + fasize;    /* Root directory start sector */
  szbfat=(fmt==FS_FAT16) ?     /* (Required FAT size) */
   fs->n_fatent * 2 : fs->n_fatent * 3 / 2 + (fs->n_fatent & 1);
 }
 if (fs->fsize < (szbfat + (SS(fs) - 1)) / SS(fs)) /* (BPB_FATSz must not be less than required) */
  return FR_NO_FILESYSTEM;

 /* Initialize cluster allocation information */
 fs->last_clust=fs->free_clust=0xFFFFFFFF;

 /* Get fsinfo if available */
 fs->fsi_flag=0x80;
 if (fmt==FS_FAT32    /* Enable FSINFO only if FAT32 and BPB_FSInfo is 1 */
  && LD_word(fs->win+BPB_FSInfo)==1
  && move_window(fs, bsect + 1)==FR_OK)
 {
  fs->fsi_flag=0;
  if (LD_word(fs->win+BS_55AA)==0xAA55 /* Load FSINFO data if available */
   && LD_dword(fs->win+FSI_LeadSig)==0x41615252
   && LD_dword(fs->win+FSI_StrucSig)==0x61417272)
  {
   fs->free_clust=LD_dword(fs->win+FSI_Free_Count);
   fs->last_clust=LD_dword(fs->win+FSI_Nxt_Free);
  }
 }
 fs->fs_type=fmt; /* FAT sub-type */
 fs->id=++Fsid; /* File system mount ID */

 return FR_OK;
}
//############################################################################//
//Check if the file/directory object is valid or not
//FR_OK(0): The object is valid, !=0: Invalid */
static FRESULT validate (void* obj)
{
 FIL *fil=(FIL*)obj; //Assuming offset of .fs and .id in the FIL/DIR structure is identical
 if (!fil || !fil->fs || !fil->fs->fs_type || fil->fs->id != fil->id) return FR_INVALID_OBJECT;
 return FR_OK;
}




/*--------------------------------------------------------------------------

   Public Functions

--------------------------------------------------------------------------*/



//Mount/Unmount a Logical Drive
FRESULT f_mount(FATFS* fs)
{
 int vol;
 FRESULT res;

 vol=0;
 if (FatFs)FatFs->fs_type=0;    //Clear old fs object

 if (fs)fs->fs_type=0;    //Clear new fs object
 FatFs=fs;     //Register new fs object
 if (!fs)return FR_OK; //Do not mount now, it will be mounted later

 res=find_volume(&fs); //Force mounted the volume
 return res;
}




/*-----------------------------------------------------------------------*/
/* Open or Create a File                                                 */
/*-----------------------------------------------------------------------*/

FRESULT f_open (
 FIL* fp,   /* Pointer to the blank file object */
 const char* path, /* Pointer to the file name */
 byte mode   /* Access mode and file open mode flags */
)
{
 FRESULT res;
 DIR dj;
 byte *dir;
 DEF_NAMEBUF;


 if (!fp) return FR_INVALID_OBJECT;
 fp->fs=0;   /* Clear file object */

 /* Get logical drive number */
 mode &= FA_READ | FA_WRITE | FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW;
 res=find_volume(&dj.fs);
 if (res==FR_OK) {
  INIT_BUF(dj);
  res=follow_path(&dj, path); /* Follow the file path */
  dir=dj.dir;
  if (res==FR_OK) {
   if (!dir) /* Default directory itself */
    res=FR_INVALID_NAME;
  }
  /* Create or Open a file */
  if (mode & (FA_CREATE_ALWAYS | FA_OPEN_ALWAYS | FA_CREATE_NEW)) {
   dword dw, cl;

   if (res != FR_OK) {     /* No file, create new */
    if (res==FR_NO_FILE)   /* There is no file to open, create a new entry */
     res=dir_register(&dj);
    mode |= FA_CREATE_ALWAYS;  /* File is created */
    dir=dj.dir;     /* New entry */
   }
   else {        /* Any object is already existing */
    if (dir[DIR_Attr] & (AM_RDO | AM_DIR)) { /* Cannot overwrite it (R/O or DIR) */
     res=FR_DENIED;
    } else {
     if (mode & FA_CREATE_NEW) /* Cannot create as new file */
      res=FR_EXIST;
    }
   }
   if (res==FR_OK && (mode & FA_CREATE_ALWAYS)) { /* Truncate it if overwrite mode */
    dw=get_fattime();    /* Created time */
    ST_dword(dir+DIR_CrtTime, dw);
    dir[DIR_Attr]=0;    /* Reset attribute */
    ST_dword(dir+DIR_FileSize, 0); /* size=0 */
    cl=ld_clust(dj.fs, dir);  /* Get start cluster */
    st_clust(dir, 0);    /* cluster=0 */
    dj.fs->wflag=1;
    if (cl) {      /* Remove the cluster chain if exist */
     dw=dj.fs->winsect;
     res=remove_chain(dj.fs, cl);
     if (res==FR_OK) {
      dj.fs->last_clust=cl - 1; /* Reuse the cluster hole */
      res=move_window(dj.fs, dw);
     }
    }
   }
  }
  else { /* Open an existing file */
   if (res==FR_OK) {     /* Follow succeeded */
    if (dir[DIR_Attr] & AM_DIR) { /* It is a directory */
     res=FR_NO_FILE;
    } else {
     if ((mode & FA_WRITE) && (dir[DIR_Attr] & AM_RDO)) /* R/O violation */
      res=FR_DENIED;
    }
   }
  }
  if (res==FR_OK) {
   if (mode & FA_CREATE_ALWAYS)  /* Set file change flag if created or overwritten */
    mode |= FA__WRITTEN;
   fp->dir_sect=dj.fs->winsect;  /* Pointer to the directory entry */
   fp->dir_ptr=dir;
  }
  
  FREE_BUF();

  if (res==FR_OK) {
   fp->flag=mode;     /* File access mode */
   fp->err=0;      /* Clear error flag */
   fp->sclust=ld_clust(dj.fs, dir); /* File start cluster */
   fp->fsize=LD_dword(dir+DIR_FileSize); /* File size */
   fp->fptr=0;      /* File pointer */
   fp->dsect=0;
#if _USE_FASTSEEK
   fp->cltbl=0;      /* Normal seek mode */
#endif
   fp->fs=dj.fs;       /* Validate file object */
   fp->id=fp->fs->id;
  }
 }

 return res;
}




/*-----------------------------------------------------------------------*/
/* Read File                                                             */
/*-----------------------------------------------------------------------*/

FRESULT f_read (
 FIL* fp,   /* Pointer to the file object */
 void* buff,  /* Pointer to data buffer */
 word btr,  /* Number of bytes to read */
 word* br  /* Pointer to number of bytes read */
)
{
 FRESULT res;
 dword clst, sect, remain;
 word rcnt, cc;
 byte csect, *rbuff=(byte*)buff;


 *br=0; /* Clear read byte counter */

 res=validate(fp);       /* Check validity */
 if (res != FR_OK) LEAVE_FF(fp->fs, res);
 if (fp->err)        /* Check error */
  LEAVE_FF(fp->fs, (FRESULT)fp->err);
 if (!(fp->flag & FA_READ))      /* Check access mode */
  LEAVE_FF(fp->fs, FR_DENIED);
 remain=fp->fsize - fp->fptr;
 if (btr > remain) btr=(word)remain;  /* Truncate btr by remaining bytes */

 for ( ;  btr;        /* Repeat until all data read */
  rbuff += rcnt, fp->fptr += rcnt, *br += rcnt, btr -= rcnt) {
  if ((fp->fptr % SS(fp->fs))==0) {  /* On the sector boundary? */
   csect=(byte)(fp->fptr / SS(fp->fs) & (fp->fs->csize - 1)); /* Sector offset in the cluster */
   if (!csect) {      /* On the cluster boundary? */
    if (fp->fptr==0) {   /* On the top of the file? */
     clst=fp->sclust;   /* Follow from the origin */
    } else {      /* Middle or end of the file */
#if _USE_FASTSEEK
     if (fp->cltbl)
      clst=clmt_clust(fp, fp->fptr); /* Get cluster# from the CLMT */
     else
#endif
      clst=get_fat(fp->fs, fp->clust); /* Follow cluster chain on the FAT */
    }
    if (clst < 2) ABORT(fp->fs, FR_int_ERR);
    if (clst==0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
    fp->clust=clst;    /* Update current cluster */
   }
   sect=clust2sect(fp->fs, fp->clust); /* Get current sector */
   if (!sect) ABORT(fp->fs, FR_int_ERR);
   sect += csect;
   cc=btr / SS(fp->fs);    /* When remaining bytes >= sector size, */
   if (cc) {       /* Read maximum contiguous sectors directly */
    if (csect + cc > fp->fs->csize) /* Clip at cluster boundary */
     cc=fp->fs->csize - csect;
    if (disk_read(rbuff, sect, cc))
     ABORT(fp->fs, FR_DISK_ERR);
#if _FS_MINIMIZE <= 2   /* Replace one of the read sectors with cached data if it contains a dirty sector */
    if (fp->fs->wflag && fp->fs->winsect - sect < cc)
     mem_cpy(rbuff + ((fp->fs->winsect - sect) * SS(fp->fs)), fp->fs->win, SS(fp->fs));
#endif
    rcnt=SS(fp->fs) * cc;   /* Number of bytes transferred */
    continue;
   }
   fp->dsect=sect;
  }
  rcnt=SS(fp->fs) - ((word)fp->fptr % SS(fp->fs)); /* Get partial sector data from sector buffer */
  if (rcnt > btr) rcnt=btr;
  if (move_window(fp->fs, fp->dsect))  /* Move sector window */
   ABORT(fp->fs, FR_DISK_ERR);
  mem_cpy(rbuff, &fp->fs->win[fp->fptr % SS(fp->fs)], rcnt); /* Pick partial sector */
 }

 LEAVE_FF(fp->fs, FR_OK);
}




/*-----------------------------------------------------------------------*/
/* Write File                                                            */
/*-----------------------------------------------------------------------*/

FRESULT f_write (
 FIL* fp,   /* Pointer to the file object */
 const void *buff, /* Pointer to the data to be written */
 word btw,   /* Number of bytes to write */
 word* bw   /* Pointer to number of bytes written */
)
{
 FRESULT res;
 dword clst, sect;
 word wcnt, cc;
 const byte *wbuff=(const byte*)buff;
 byte csect;


 *bw=0; /* Clear write byte counter */

 res=validate(fp);      /* Check validity */
 if (res != FR_OK) LEAVE_FF(fp->fs, res);
 if (fp->err)       /* Check error */
  LEAVE_FF(fp->fs, (FRESULT)fp->err);
 if (!(fp->flag & FA_WRITE))    /* Check access mode */
  LEAVE_FF(fp->fs, FR_DENIED);
 if (fp->fptr + btw < fp->fptr) btw=0; /* File size cannot reach 4GB */

 for ( ;  btw;       /* Repeat until all data written */
  wbuff += wcnt, fp->fptr += wcnt, *bw += wcnt, btw -= wcnt) {
  if ((fp->fptr % SS(fp->fs))==0) { /* On the sector boundary? */
   csect=(byte)(fp->fptr / SS(fp->fs) & (fp->fs->csize - 1)); /* Sector offset in the cluster */
   if (!csect) {     /* On the cluster boundary? */
    if (fp->fptr==0) {  /* On the top of the file? */
     clst=fp->sclust;  /* Follow from the origin */
     if (clst==0)   /* When no cluster is allocated, */
      fp->sclust=clst=create_chain(fp->fs, 0); /* Create a new cluster chain */
    } else {     /* Middle or end of the file */
#if _USE_FASTSEEK
     if (fp->cltbl)
      clst=clmt_clust(fp, fp->fptr); /* Get cluster# from the CLMT */
     else
#endif
      clst=create_chain(fp->fs, fp->clust); /* Follow or stretch cluster chain on the FAT */
    }
    if (clst==0) break;  /* Could not allocate a new cluster (disk full) */
    if (clst==1) ABORT(fp->fs, FR_int_ERR);
    if (clst==0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
    fp->clust=clst;   /* Update current cluster */
   }
   if (fp->fs->winsect==fp->dsect && sync_window(fp->fs)) /* Write-back sector cache */
    ABORT(fp->fs, FR_DISK_ERR);
   sect=clust2sect(fp->fs, fp->clust); /* Get current sector */
   if (!sect) ABORT(fp->fs, FR_int_ERR);
   sect += csect;
   cc=btw / SS(fp->fs);   /* When remaining bytes >= sector size, */
   if (cc) {      /* Write maximum contiguous sectors directly */
    if (csect + cc > fp->fs->csize) /* Clip at cluster boundary */
     cc=fp->fs->csize - csect;
    if (disk_write(wbuff, sect, cc))
     ABORT(fp->fs, FR_DISK_ERR);
#if _FS_MINIMIZE <= 2
    if (fp->fs->winsect - sect < cc) { /* Refill sector cache if it gets invalidated by the direct write */
     mem_cpy(fp->fs->win, wbuff + ((fp->fs->winsect - sect) * SS(fp->fs)), SS(fp->fs));
     fp->fs->wflag=0;
    }
#endif
    wcnt=SS(fp->fs) * cc;  /* Number of bytes transferred */
    continue;
   }
   if (fp->fptr >= fp->fsize) { /* Avoid silly cache filling at growing edge */
    if (sync_window(fp->fs)) ABORT(fp->fs, FR_DISK_ERR);
    fp->fs->winsect=sect;
   }
   fp->dsect=sect;
  }
  wcnt=SS(fp->fs) - ((word)fp->fptr % SS(fp->fs));/* Put partial sector into file I/O buffer */
  if (wcnt > btw) wcnt=btw;
  if (move_window(fp->fs, fp->dsect)) /* Move sector window */
   ABORT(fp->fs, FR_DISK_ERR);
  mem_cpy(&fp->fs->win[fp->fptr % SS(fp->fs)], wbuff, wcnt); /* Fit partial sector */
  fp->fs->wflag=1;
 }

 if (fp->fptr > fp->fsize) fp->fsize=fp->fptr; /* Update file size if needed */
 fp->flag |= FA__WRITTEN;      /* Set file change flag */

 LEAVE_FF(fp->fs, FR_OK);
}




/*-----------------------------------------------------------------------*/
/* Synchronize the File                                                  */
/*-----------------------------------------------------------------------*/

FRESULT f_sync (
 FIL* fp  /* Pointer to the file object */
)
{
 FRESULT res;
 dword tm;
 byte *dir;


 res=validate(fp);     /* Check validity of the object */
 if (res==FR_OK) {
  if (fp->flag & FA__WRITTEN) { /* Has the file been written? */
   /* Write-back dirty buffer */
   /* Update the directory entry */
   res=move_window(fp->fs, fp->dir_sect);
   if (res==FR_OK) {
    dir=fp->dir_ptr;
    dir[DIR_Attr] |= AM_ARC;     /* Set archive bit */
    ST_dword(dir+DIR_FileSize, fp->fsize);  /* Update file size */
    st_clust(dir, fp->sclust);     /* Update start cluster */
    tm=get_fattime();       /* Update updated time */
    ST_dword(dir+DIR_WrtTime, tm);
    ST_word(dir+DIR_LstAccDate, 0);
    fp->flag &= ~FA__WRITTEN;
    fp->fs->wflag=1;
    res=sync_fs(fp->fs);
   }
  }
 }

 LEAVE_FF(fp->fs, res);
}
//############################################################################//
//Close File
FRESULT f_close (FIL *fp)
{
 FRESULT res;

 res=f_sync(fp);     //Flush cached data
 if (res==FR_OK){
  res=validate(fp); 
  if (res==FR_OK)fp->fs=0;    //Invalidate file object
 }
 return res;
}



#if _FS_MINIMIZE <= 2
/*-----------------------------------------------------------------------*/
/* Seek File R/W Pointer                                                 */
/*-----------------------------------------------------------------------*/

FRESULT f_lseek (
 FIL* fp,  /* Pointer to the file object */
 dword ofs  /* File pointer from top of file */
)
{
 FRESULT res;


 res=validate(fp);     /* Check validity of the object */
 if (res != FR_OK) LEAVE_FF(fp->fs, res);
 if (fp->err)      /* Check error */
  LEAVE_FF(fp->fs, (FRESULT)fp->err);

#if _USE_FASTSEEK
 if (fp->cltbl) { /* Fast seek */
  dword cl, pcl, ncl, tcl, dsc, tlen, ulen, *tbl;

  if (ofs==CREATE_LINKMAP) { /* Create CLMT */
   tbl=fp->cltbl;
   tlen=*tbl++; ulen=2; /* Given table size and required table size */
   cl=fp->sclust;   /* Top of the chain */
   if (cl) {
    do {
     /* Get a fragment */
     tcl=cl; ncl=0; ulen += 2; /* Top, length and used items */
     do {
      pcl=cl; ncl++;
      cl=get_fat(fp->fs, cl);
      if (cl <= 1) ABORT(fp->fs, FR_int_ERR);
      if (cl==0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
     } while (cl==pcl + 1);
     if (ulen <= tlen) {  /* Store the length and top of the fragment */
      *tbl++=ncl; *tbl++=tcl;
     }
    } while (cl < fp->fs->n_fatent); /* Repeat until end of chain */
   }
   *fp->cltbl=ulen; /* Number of items used */
   if (ulen <= tlen)
    *tbl=0;  /* Terminate table */
   else
    res=FR_NOT_ENOUGH_CORE; /* Given table size is smaller than required */

  } else {      /* Fast seek */
   if (ofs > fp->fsize)  /* Clip offset at the file size */
    ofs=fp->fsize;
   fp->fptr=ofs;    /* Set file pointer */
   if (ofs) {
    fp->clust=clmt_clust(fp, ofs - 1);
    dsc=clust2sect(fp->fs, fp->clust);
    if (!dsc) ABORT(fp->fs, FR_int_ERR);
    dsc += (ofs - 1) / SS(fp->fs) & (fp->fs->csize - 1);
    if (fp->fptr % SS(fp->fs) && dsc != fp->dsect) { /* Refill sector cache if needed */
     fp->dsect=dsc;
    }
   }
  }
 } else
#endif

 /* Normal Seek */
 {
  dword clst, bcs, nsect, ifptr;

  if (ofs > fp->fsize     /* In read-only mode, clip offset with the file size */
    && !(fp->flag & FA_WRITE)
   ) ofs=fp->fsize;

  ifptr=fp->fptr;
  fp->fptr=nsect=0;
  if (ofs) {
   bcs=(dword)fp->fs->csize * SS(fp->fs); /* Cluster size (byte) */
   if (ifptr > 0 &&
    (ofs - 1) / bcs >= (ifptr - 1) / bcs) { /* When seek to same or following cluster, */
    fp->fptr=(ifptr - 1) & ~(bcs - 1); /* start from the current cluster */
    ofs -= fp->fptr;
    clst=fp->clust;
   } else {         /* When seek to back cluster, */
    clst=fp->sclust;      /* start from the first cluster */
    if (clst==0) {      /* If no cluster chain, create a new chain */
     clst=create_chain(fp->fs, 0);
     if (clst==1) ABORT(fp->fs, FR_int_ERR);
     if (clst==0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
     fp->sclust=clst;
    }
    fp->clust=clst;
   }
   if (clst != 0) {
    while (ofs > bcs) {      /* Cluster following loop */
     if (fp->flag & FA_WRITE) {   /* Check if in write mode or not */
      clst=create_chain(fp->fs, clst); /* Force stretch if in write mode */
      if (clst==0) {    /* When disk gets full, clip file size */
       ofs=bcs; break;
      }
     } else
      clst=get_fat(fp->fs, clst); /* Follow cluster chain if not in write mode */
     if (clst==0xFFFFFFFF) ABORT(fp->fs, FR_DISK_ERR);
     if (clst <= 1 || clst >= fp->fs->n_fatent) ABORT(fp->fs, FR_int_ERR);
     fp->clust=clst;
     fp->fptr += bcs;
     ofs -= bcs;
    }
    fp->fptr += ofs;
    if (ofs % SS(fp->fs)) {
     nsect=clust2sect(fp->fs, clst); /* Current sector */
     if (!nsect) ABORT(fp->fs, FR_int_ERR);
     nsect += ofs / SS(fp->fs);
    }
   }
  }
  if (fp->fptr % SS(fp->fs) && nsect != fp->dsect) { /* Fill sector cache if needed */
   fp->dsect=nsect;
  }
  if (fp->fptr > fp->fsize) {   /* Set file change flag if the file size is extended */
   fp->fsize=fp->fptr;
   fp->flag |= FA__WRITTEN;
  }
 }

 LEAVE_FF(fp->fs, res);
}



#if _FS_MINIMIZE <= 1
/*-----------------------------------------------------------------------*/
/* Create a Directory Object                                             */
/*-----------------------------------------------------------------------*/

FRESULT f_opendir (
 DIR* dp,   /* Pointer to directory object to create */
 const char* path /* Pointer to the directory path */
)
{
 FRESULT res;
 FATFS* fs;
 DEF_NAMEBUF;


 if (!dp) return FR_INVALID_OBJECT;

 /* Get logical drive number */
 res=find_volume(&fs);
 if (res==FR_OK) {
  dp->fs=fs;
  INIT_BUF(*dp);
  res=follow_path(dp, path);   /* Follow the path to the directory */
  FREE_BUF();
  if (res==FR_OK) {      /* Follow completed */
   if (dp->dir) {      /* It is not the origin directory itself */
    if (dp->dir[DIR_Attr] & AM_DIR) /* The object is a sub directory */
     dp->sclust=ld_clust(fs, dp->dir);
    else       /* The object is a file */
     res=FR_NO_PATH;
   }
   if (res==FR_OK) {
    dp->id=fs->id;
    res=dir_sdi(dp, 0);   /* Rewind directory */
   }
  }
  if (res==FR_NO_FILE) res=FR_NO_PATH;
 }
 if (res != FR_OK) dp->fs=0;  /* Invalidate the directory object if function faild */

 LEAVE_FF(fs, res);
}




/*-----------------------------------------------------------------------*/
/* Close Directory                                                       */
/*-----------------------------------------------------------------------*/

FRESULT f_closedir (
 DIR *dp  /* Pointer to the directory object to be closed */
)
{
 FRESULT res;


 res=validate(dp);
 if (res==FR_OK) {
   dp->fs=0;    /* Invalidate directory object */
 }
 return res;
}




/*-----------------------------------------------------------------------*/
/* Read Directory Entries in Sequence                                    */
/*-----------------------------------------------------------------------*/

FRESULT f_readdir (
 DIR* dp,   /* Pointer to the open directory object */
 FILINFO* fno  /* Pointer to file information to return */
)
{
 FRESULT res;
 DEF_NAMEBUF;


 res=validate(dp);      /* Check validity of the object */
 if (res==FR_OK) {
  if (!fno) {
   res=dir_sdi(dp, 0);   /* Rewind the directory object */
  } else {
   INIT_BUF(*dp);
   res=dir_read(dp, 0);   /* Read an item */
   if (res==FR_NO_FILE) {  /* Reached end of directory */
    dp->sect=0;
    res=FR_OK;
   }
   if (res==FR_OK) {    /* A valid entry is found */
    get_fileinfo(dp, fno);  /* Get the object information */
    res=dir_next(dp, 0);  /* Increment index for next */
    if (res==FR_NO_FILE) {
     dp->sect=0;
     res=FR_OK;
    }
   }
   FREE_BUF();
  }
 }

 LEAVE_FF(dp->fs, res);
}



#if _FS_MINIMIZE==0
/*-----------------------------------------------------------------------*/
/* Get File Status                                                       */
/*-----------------------------------------------------------------------*/

FRESULT f_stat (
 const char* path, /* Pointer to the file path */
 FILINFO* fno  /* Pointer to file information to return */
)
{
 FRESULT res;
 DIR dj;
 DEF_NAMEBUF;


 /* Get logical drive number */
 res=find_volume(&dj.fs);
 if (res==FR_OK) {
  INIT_BUF(dj);
  res=follow_path(&dj, path); /* Follow the file path */
  if (res==FR_OK) {    /* Follow completed */
   if (dj.dir) {  /* Found an object */
    if (fno) get_fileinfo(&dj, fno);
   } else {   /* It is root directory */
    res=FR_INVALID_NAME;
   }
  }
  FREE_BUF();
 }

 return res;
}



/*-----------------------------------------------------------------------*/
/* Get Number of Free Clusters                                           */
/*-----------------------------------------------------------------------*/

FRESULT f_getfree (
 dword* nclst,  /* Pointer to a variable to return number of free clusters */
 FATFS** fatfs  /* Pointer to return pointer to corresponding file system object */
)
{
 FRESULT res;
 FATFS *fs;
 dword n, clst, sect, stat;
 word i;
 byte fat, *p;


 /* Get logical drive number */
 res=find_volume(fatfs);
 fs=*fatfs;
 if (res==FR_OK) {
  /* If free_clust is valid, return it without full cluster scan */
  if (fs->free_clust <= fs->n_fatent - 2) {
   *nclst=fs->free_clust;
  } else {
   /* Get number of free clusters */
   fat=fs->fs_type;
   n=0;
   if (fat==FS_FAT12) {
    clst=2;
    do {
     stat=get_fat(fs, clst);
     if (stat==0xFFFFFFFF) { res=FR_DISK_ERR; break; }
     if (stat==1) { res=FR_int_ERR; break; }
     if (stat==0) n++;
    } while (++clst < fs->n_fatent);
   } else {
    clst=fs->n_fatent;
    sect=fs->fatbase;
    i=0; p=0;
    do {
     if (!i) {
      res=move_window(fs, sect++);
      if (res != FR_OK) break;
      p=fs->win;
      i=SS(fs);
     }
     if (fat==FS_FAT16) {
      if (LD_word(p)==0) n++;
      p += 2; i -= 2;
     } else {
      if ((LD_dword(p) & 0x0FFFFFFF)==0) n++;
      p += 4; i -= 4;
     }
    } while (--clst);
   }
   fs->free_clust=n;
   fs->fsi_flag |= 1;
   *nclst=n;
  }
 }
 LEAVE_FF(fs, res);
}




/*-----------------------------------------------------------------------*/
/* Truncate File                                                         */
/*-----------------------------------------------------------------------*/

FRESULT f_truncate (
 FIL* fp  /* Pointer to the file object */
)
{
 FRESULT res;
 dword ncl;


 res=validate(fp);      /* Check validity of the object */
 if (res==FR_OK) {
  if (fp->err) {      /* Check error */
   res=(FRESULT)fp->err;
  } else {
   if (!(fp->flag & FA_WRITE))  /* Check access mode */
    res=FR_DENIED;
  }
 }
 if (res==FR_OK) {
  if (fp->fsize > fp->fptr) {
   fp->fsize=fp->fptr; /* Set file size to current R/W point */
   fp->flag |= FA__WRITTEN;
   if (fp->fptr==0) { /* When set file size to zero, remove entire cluster chain */
    res=remove_chain(fp->fs, fp->sclust);
    fp->sclust=0;
   } else {    /* When truncate a part of the file, remove remaining clusters */
    ncl=get_fat(fp->fs, fp->clust);
    res=FR_OK;
    if (ncl==0xFFFFFFFF) res=FR_DISK_ERR;
    if (ncl==1) res=FR_int_ERR;
    if (res==FR_OK && ncl < fp->fs->n_fatent) {
     res=put_fat(fp->fs, fp->clust, 0x0FFFFFFF);
     if (res==FR_OK) res=remove_chain(fp->fs, ncl);
    }
   }
  }
  if (res != FR_OK) fp->err=(FRESULT)res;
 }

 LEAVE_FF(fp->fs, res);
}




/*-----------------------------------------------------------------------*/
/* Delete a File or Directory                                            */
/*-----------------------------------------------------------------------*/

FRESULT f_unlink (
 const char* path  /* Pointer to the file or directory path */
)
{
 FRESULT res;
 DIR dj, sdj;
 byte *dir;
 dword dclst;
 DEF_NAMEBUF;


 /* Get logical drive number */
 res=find_volume(&dj.fs);
 if (res==FR_OK) {
  INIT_BUF(dj);
  res=follow_path(&dj, path);  /* Follow the file path */
  if (res==FR_OK) {     /* The object is accessible */
   dir=dj.dir;
   if (!dir) {
    res=FR_INVALID_NAME;  /* Cannot remove the start directory */
   } else {
    if (dir[DIR_Attr] & AM_RDO)
     res=FR_DENIED;  /* Cannot remove R/O object */
   }
   dclst=ld_clust(dj.fs, dir);
   if (res==FR_OK && (dir[DIR_Attr] & AM_DIR)) { /* Is it a sub-dir? */
    if (dclst < 2) {
     res=FR_int_ERR;
    } else {
     mem_cpy(&sdj, &dj, sizeof (DIR)); /* Check if the sub-directory is empty or not */
     sdj.sclust=dclst;
     res=dir_sdi(&sdj, 2);  /* Exclude dot entries */
     if (res==FR_OK) {
      res=dir_read(&sdj, 0); /* Read an item */
      if (res==FR_OK  /* Not empty directory */
      ) res=FR_DENIED;
      if (res==FR_NO_FILE) res=FR_OK; /* Empty */
     }
    }
   }
   if (res==FR_OK) {
    res=dir_remove(&dj);  /* Remove the directory entry */
    if (res==FR_OK) {
     if (dclst)    /* Remove the cluster chain if exist */
      res=remove_chain(dj.fs, dclst);
     if (res==FR_OK) res=sync_fs(dj.fs);
    }
   }
  }
  FREE_BUF();
 }

 LEAVE_FF(dj.fs, res);
}




/*-----------------------------------------------------------------------*/
/* Create a Directory                                                    */
/*-----------------------------------------------------------------------*/

FRESULT f_mkdir(const char* path)
{
 FRESULT res;
 DIR dj;
 byte *dir, n;
 dword dsc, dcl, pcl, tm=get_fattime();
 DEF_NAMEBUF;


 /* Get logical drive number */
 res=find_volume(&dj.fs);
 if (res==FR_OK) {
  INIT_BUF(dj);
  res=follow_path(&dj, path);   /* Follow the file path */
  if (res==FR_OK) res=FR_EXIST;  /* Any object with same name is already existing */
  if (res==FR_NO_FILE) {    /* Can create a new directory */
   dcl=create_chain(dj.fs, 0);  /* Allocate a cluster for the new directory table */
   res=FR_OK;
   if (dcl==0) res=FR_DENIED;  /* No space to allocate a new cluster */
   if (dcl==1) res=FR_int_ERR;
   if (dcl==0xFFFFFFFF) res=FR_DISK_ERR;
   if (res==FR_OK)     /* Flush FAT */
    res=sync_window(dj.fs);
   if (res==FR_OK) {     /* Initialize the new directory table */
    dsc=clust2sect(dj.fs, dcl);
    dir=dj.fs->win;
    mem_set(dir, 0, SS(dj.fs));
    mem_set(dir+DIR_Name, ' ', 11); /* Create "." entry */
    dir[DIR_Name]='.';
    dir[DIR_Attr]=AM_DIR;
    ST_dword(dir+DIR_WrtTime, tm);
    st_clust(dir, dcl);
    mem_cpy(dir+SZ_DIR, dir, SZ_DIR);  /* Create ".." entry */
    dir[SZ_DIR+1]='.'; pcl=dj.sclust;
    if (dj.fs->fs_type==FS_FAT32 && pcl==dj.fs->dirbase)
     pcl=0;
    st_clust(dir+SZ_DIR, pcl);
    for (n=dj.fs->csize; n; n--) { /* Write dot entries and clear following sectors */
     dj.fs->winsect=dsc++;
     dj.fs->wflag=1;
     res=sync_window(dj.fs);
     if (res != FR_OK) break;
     mem_set(dir, 0, SS(dj.fs));
    }
   }
   if (res==FR_OK) res=dir_register(&dj); /* Register the object to the directoy */
   if (res != FR_OK) {
    remove_chain(dj.fs, dcl);   /* Could not register, remove cluster chain */
   } else {
    dir=dj.dir;
    dir[DIR_Attr]=AM_DIR;    /* Attribute */
    ST_dword(dir+DIR_WrtTime, tm);  /* Created time */
    st_clust(dir, dcl);     /* Table start cluster */
    dj.fs->wflag=1;
    res=sync_fs(dj.fs);
   }
  }
  FREE_BUF();
 }

 LEAVE_FF(dj.fs, res);
}




/*-----------------------------------------------------------------------*/
/* Change Attribute                                                      */
/*-----------------------------------------------------------------------*/

FRESULT f_chmod (
 const char* path, /* Pointer to the file path */
 byte value,   /* Attribute bits */
 byte mask   /* Attribute mask to change */
)
{
 FRESULT res;
 DIR dj;
 byte *dir;
 DEF_NAMEBUF;


 /* Get logical drive number */
 res=find_volume(&dj.fs);
 if (res==FR_OK) {
  INIT_BUF(dj);
  res=follow_path(&dj, path);  /* Follow the file path */
  FREE_BUF();
  if (res==FR_OK) {
   dir=dj.dir;
   if (!dir) {      /* Is it a root directory? */
    res=FR_INVALID_NAME;
   } else {      /* File or sub directory */
    mask &= AM_RDO|AM_HID|AM_SYS|AM_ARC; /* Valid attribute mask */
    dir[DIR_Attr]=(value & mask) | (dir[DIR_Attr] & (byte)~mask); /* Apply attribute change */
    dj.fs->wflag=1;
    res=sync_fs(dj.fs);
   }
  }
 }

 LEAVE_FF(dj.fs, res);
}




/*-----------------------------------------------------------------------*/
/* Change Timestamp                                                      */
/*-----------------------------------------------------------------------*/

FRESULT f_utime (
 const char* path, /* Pointer to the file/directory name */
 const FILINFO* fno /* Pointer to the time stamp to be set */
)
{
 FRESULT res;
 DIR dj;
 byte *dir;
 DEF_NAMEBUF;


 /* Get logical drive number */
 res=find_volume(&dj.fs);
 if (res==FR_OK) {
  INIT_BUF(dj);
  res=follow_path(&dj, path); /* Follow the file path */
  FREE_BUF();
  if (res==FR_OK) {
   dir=dj.dir;
   if (!dir) {     /* Root directory */
    res=FR_INVALID_NAME;
   } else {     /* File or sub-directory */
    ST_word(dir+DIR_WrtTime, fno->ftime);
    ST_word(dir+DIR_WrtDate, fno->fdate);
    dj.fs->wflag=1;
    res=sync_fs(dj.fs);
   }
  }
 }

 LEAVE_FF(dj.fs, res);
}




/*-----------------------------------------------------------------------*/
/* Rename File/Directory                                                 */
/*-----------------------------------------------------------------------*/

FRESULT f_rename (
 const char* path_old, /* Pointer to the old name */
 const char* path_new /* Pointer to the new name */
)
{
 FRESULT res;
 DIR djo, djn;
 byte buf[21], *dir;
 dword dw;
 DEF_NAMEBUF;


 /* Get logical drive number of the source object */
 res=find_volume(&djo.fs);
 if (res==FR_OK) {
  djn.fs=djo.fs;
  INIT_BUF(djo);
  res=follow_path(&djo, path_old);  /* Check old object */
  if (res==FR_OK) {      /* Old object is found */
   if (!djo.dir) {      /* Is root dir? */
    res=FR_NO_FILE;
   } else {
    mem_cpy(buf, djo.dir+DIR_Attr, 21);  /* Save the object information except for name */
    mem_cpy(&djn, &djo, sizeof (DIR));  /* Check new object */
    res=follow_path(&djn, path_new);
    if (res==FR_OK) res=FR_EXIST;  /* The new object name is already existing */
    if (res==FR_NO_FILE) {     /* Is it a valid path and no name collision? */
/* Start critical section that any interruption can cause a cross-link */
     res=dir_register(&djn);   /* Register the new entry */
     if (res==FR_OK) {
      dir=djn.dir;     /* Copy object information except for name */
      mem_cpy(dir+13, buf+2, 19);
      dir[DIR_Attr]=buf[0] | AM_ARC;
      djo.fs->wflag=1;
      if (djo.sclust != djn.sclust && (dir[DIR_Attr] & AM_DIR)) {  /* Update .. entry in the directory if needed */
       dw=clust2sect(djo.fs, ld_clust(djo.fs, dir));
       if (!dw) {
        res=FR_int_ERR;
       } else {
        res=move_window(djo.fs, dw);
        dir=djo.fs->win+SZ_DIR; /* .. entry */
        if (res==FR_OK && dir[1]=='.') {
         dw=(djo.fs->fs_type==FS_FAT32 && djn.sclust==djo.fs->dirbase) ? 0 : djn.sclust;
         st_clust(dir, dw);
         djo.fs->wflag=1;
        }
       }
      }
      if (res==FR_OK) {
       res=dir_remove(&djo);  /* Remove old entry */
       if (res==FR_OK)
        res=sync_fs(djo.fs);
      }
     }
/* End critical section */
    }
   }
  }
  FREE_BUF();
 }

 LEAVE_FF(djo.fs, res);
}

#endif /* _FS_MINIMIZE==0 */
#endif /* _FS_MINIMIZE <= 1 */
#endif /* _FS_MINIMIZE <= 2 */
//############################################################################//